from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import IntEnum
from functools import cached_property
from ipaddress import IPv4Address, IPv6Address
from socket import htonl
from struct import pack, unpack
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystem import fsutil
from dissect.target.helpers.utils import StrEnum
from dissect.target.plugin import Plugin, internal

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.target import Target


def parse_ip(addr: str | int, version: int = 4) -> IPv6Address | IPv4Address:
    """Convert ``/proc/net`` IPv4 or IPv6 hex address into their standard IP notation."""

    if version == 6:
        addr = unpack("!LLLL", bytes.fromhex(addr))
        return IPv6Address(pack("@IIII", *addr))

    if isinstance(addr, int):
        return IPv4Address(htonl(addr))

    addr = int(addr, 16)
    return IPv4Address(htonl(addr))


# Labels are borrowed from https://www.kernel.org/doc/Documentation/networking/proc_net_tcp.txt
@dataclass(order=True)
class NetSocket:
    sl: str  # number of entry
    local_address: str  # local IPv4 or IPv6 address
    rem_address: str  # remote IPv4 or IPv6 address
    state: str  # connection state
    tx_rx_queue: str  # transmit and receive-queue
    tr_tm_when: str  # timer_active and number of jiffies until timer expires
    restansmit: str  # number of unrecovered RTO timeouts
    uid: int  # uid
    timeout: str  # unanswered 0-window probes
    inode: int  # inode
    ref: str | None = None  # socket reference count
    pointer: str | None = None  # location of socket in memory
    drops: str | None = None  # retransmit timeout
    predicted_tick: str | None = None  # predicted tick of soft slock (delayed ACK control data)
    ack_pingpong: str | None = None  # ack.quick<<1|ack.pingpong
    congestion_window: str | None = None  # sending congestion window
    size_threshold: str | None = None  # slow start size threshhold or -f if the threshold is >= 0xFFFF

    # Values parsed from raw values listed above.
    protocol_string: str | None = None
    local_ip: str | None = None  # parsed value from local_address
    local_port: int | None = None  # parsed value from local_address
    remote_ip: str | None = None  # parsed value from rem_address
    remote_port: int | None = None  # parsed value from rem_address
    state_string: str | None = None  # parsed value from state
    owner: str | None = None  # resolved owner name of the socket, else "0".
    rx_queue: int | None = None  # parsed value from tx_rx_queue
    tx_queue: int | None = None  # parsed value from tx_rx_queue
    pid: int | None = None  # pid of the socket
    name: str | None = None  # process name associated to the socket
    cmdline: str | None = None  # process cmdline associated to the socket

    @classmethod
    def from_line(cls, line: str, ip_vers: int = 4) -> Self:
        socket = cls(*line.split())

        socket.uid = int(socket.uid)
        socket.inode = int(socket.inode)
        socket.tx_queue, socket.rx_queue = [int(queue, 16) for queue in socket.tx_rx_queue.split(":")]
        socket.local_ip, socket.local_port = socket.local_address.split(":")
        socket.remote_ip, socket.remote_port = socket.rem_address.split(":")

        socket.local_ip = parse_ip(socket.local_ip, ip_vers)
        socket.local_port = int(socket.local_port, 16)
        socket.remote_ip = parse_ip(socket.remote_ip, ip_vers)
        socket.remote_port = int(socket.remote_port, 16)

        return socket


@dataclass(order=True)
class UnixSocket:
    num: str  # the kernel table slot number
    ref: int  # the number of users of the socket
    protocol: int  # currently always 0. "Unix"
    flags: str  # the internal kernel flags holding the status of the socket
    type: int  # the socket type: 1 for SOCK_STREAM, 2 for SOCK_DGRAM and 5 for SOCK_SEQPACKET sockets
    state: int  # the internal state of the socket
    inode: int  # the inode number of the socket. the inode is commonly refered to as port in tools as ss and netstat
    path: str | None = None  # sockets in the abstract namespace are included in the list,
    # and are shown with a Path that commences with the character '@'

    # Values parsed from raw values listed above.
    state_string: str | None = None
    stream_type_string: str | None = None
    protocol_string: str = "unix"

    @classmethod
    def from_line(cls, line: str) -> Self:
        socket = cls(*line.split())

        socket.type = int(socket.type)
        socket.protocol = int(socket.protocol)
        socket.state = int(socket.state, 16)
        socket.ref = int(socket.ref, 16)
        socket.inode = int(socket.inode)

        return socket


@dataclass(order=True)
class PacketSocket:
    sk: int  # socket number
    ref: int  # the number of processes using this socket
    type: int  # type of the socket
    protocol: int  # protocol used by the socket to capture ie. 0003 is ETH_P_ALL
    iface: int  # network interface index
    r: int  # number of bytes that have been received by the socket and are waiting to be processed
    rmem: int  # receive window memory
    user: int  # uid from the owner of the socket
    inode: int  # inode

    # Values parsed from raw values listed above
    pid: int | None = None  # pid of the socket
    name: str | None = None  # process name associated to the socket
    cmdline: str | None = None  # process cmdline associated to the socket
    protocol_type: int | None = None  # value parsed from protocol field
    owner: str | None = None  # resolved owner from user (uid) field
    protocol_string: str = "packet"

    @classmethod
    def from_line(cls, line: str) -> Self:
        parts = line.split()
        return cls(*[int(parts[0], 16), *list(map(int, parts[1:]))])


@dataclass
class Environ:
    variable: str
    contents: str


class ProcessStateEnum(StrEnum):
    R = "Running"  # Running
    I = "Idle"  # Idle # noqa: E741
    S = "Sleeping"  # Sleeping in an interruptible wait
    D = "Waiting"  # Waiting in uninterruptible disk sleep
    Z = "Zombie"  # Zombie
    T = "Stopped"  # Stopped (on a signal) or (before Linux 2.6.33) trace stopped
    t = "Tracing"  # Tracing stop (Linux 2.6.33 onward)
    X = "Dead"  # Dead (from Linux 2.6.0 onward)
    x = "Dead"  # Dead (Linux 2.6.33 to 3.13 only)
    K = "Wakekill"  # Wakekill (Linux 2.6.33 to 3.13 only)
    W = "Waking"  # Waking (Linux 2.6.33 to 3.13 only)
    P = "Parked"  # Parked (Linux 3.9 to 3.13 only)
    N = "None"  # Sentinel value in-case a process' state file is not present


PROC_STAT_NAMES = [
    "pid",
    "comm",  # Process name
    "state",
    "ppid",
    "pgrp",
    "session",
    "tty_nr",
    "tpgid",
    "flags",
    "minflt",
    "cminflt",
    "majflt",
    "cmajflt",
    "utime",
    "stime",
    "cutime",
    "cstime",
    "priority",
    "nice",
    "num_threads",
    "itrealvalue",
    "starttime",
    "vsize",
    "rss",
    "rsslim",
    "startcode",
    "endcode",
    "startstack",
    "kstkesp",
    "kstkeip",
    "signal",
    "blocked",
    "sigignore",
    "sigcatch",
    "wchan",
    "nswap",
    "cnswap",
    "exit_signal",
    "processor",
    "rt_priority",
    "policy",
    "delayacct_blkio_ticks",
    "guest_time",
    "cguest_time",
    "start_data",
    "end_data",
    "start_brk",
    "arg_start",
    "arg_end",
    "env_start",
    "env_end",
    "exit_code",
]


class Sockets:
    class PacketProtocolTypes(IntEnum):
        ETH_P_802_3 = 0x0001  # Dummy type for 802.3 frames
        ETH_P_AX25 = 0x0002  # Dummy protocol id for AX.25
        ETH_P_ALL = 0x0003  # Every packet (be careful!!!)
        ETH_P_802_2 = 0x0004  # 802.2 frames
        ETH_P_SNAP = 0x0005  # Internal only
        ETH_P_DDCMP = 0x0006  # DEC DDCMP: Internal only
        ETH_P_WAN_PPP = 0x0007  # Dummy type for WAN PPP frames
        ETH_P_PPP_MP = 0x0008  # Dummy type for PPP MP frames
        ETH_P_LOCALTALK = 0x0009  # Localtalk pseudo type
        ETH_P_CAN = 0x000C  # Controller Area Network
        ETH_P_PPPTALK = 0x0010  # Dummy type for Atalk over PPP
        ETH_P_TR_802_2 = 0x0011  # 802.2 frames
        ETH_P_MOBITEX = 0x0015  # Mobitex (kaz@cafe.net)
        ETH_P_CONTROL = 0x0016  # Card specific control frames
        ETH_P_IRDA = 0x0017  # Linux-IrDA
        ETH_P_ECONET = 0x0018  # Acorn Econet
        ETH_P_HDLC = 0x0019  # HDLC frames
        ETH_P_ARCNET = 0x001A  # 1A for ArcNet :-)
        ETH_P_DSA = 0x001B  # Distributed Switch Arch.
        ETH_P_TRAILER = 0x001C  # Trailer switch tagging
        ETH_P_PHONET = 0x00F5  # Nokia Phonet frames
        ETH_P_IEEE802154 = 0x00F6  # IEEE802.15.4 frame

    class SocketStreamType(IntEnum):
        STREAM = 1
        DGRAM = 2
        SEQPACKET = 5

    class SocketStateType(IntEnum):
        LISTENING = 1
        CONNECTED = 3

    class TCPStates(IntEnum):
        DUMMY = 0
        ESTABLISHED = 1
        SYN_SENT = 2
        SYN_RECV = 3
        FIN_WAIT1 = 4
        FIN_WAIT2 = 5
        TIME_WAIT = 6
        CLOSE = 7
        CLOSE_WAIT = 8
        LAST_ACK = 9
        LISTEN = 10
        CLOSING = 11
        NEW_SYN_RECV = 12
        MAX_STATES = 13

    class UDPStates(IntEnum):
        DUMMY = 0
        ESTABLISHED = 1
        LISTEN = 7

    def __init__(self, target: Target):
        self.target = target

    def packet(self) -> Iterator[PacketSocket]:
        """Yield parsed ``/proc/net/packet`` entries."""
        yield from self._parse_packet_sockets()

    def raw(self) -> Iterator[NetSocket]:
        """Yield parsed ``/proc/net/raw`` entries."""
        yield from self._parse_net_sockets("raw")

    def raw6(self) -> Iterator[NetSocket]:
        """Yield parsed ``/proc/net/raw6`` entries."""
        yield from self._parse_net_sockets("raw", 6)

    def tcp6(self) -> Iterator[NetSocket]:
        """Yield parsed ``/proc/net/tcp6`` entries."""
        yield from self._parse_net_sockets("tcp", 6)

    def tcp(self) -> Iterator[NetSocket]:
        """Yield parsed ``/proc/net/tcp`` entries."""
        yield from self._parse_net_sockets("tcp")

    def udp(self) -> Iterator[NetSocket]:
        """Yield parsed ``/proc/net/upd`` entries."""
        yield from self._parse_net_sockets("udp")

    def udp6(self) -> Iterator[NetSocket]:
        """Yield parsed ``/proc/net/udp6`` entries."""
        yield from self._parse_net_sockets("udp", 6)

    def unix(self) -> Iterator[UnixSocket]:
        """Yield parsed ``/proc/net/unix`` entries."""
        yield from self._parse_unix_sockets()

    def _parse_net_sockets(self, protocol: str = "tcp", version: int = 4) -> Iterator[NetSocket]:
        """Internal function to parse ``/proc/net/{tcp(6),udp(6), raw(6)}`` entries.

        Args:
            protocol: The protocol in ``/proc/net/`` to parse entries from.
            version: The version of the protocol to parse entries from.
        """

        entry = self.target.fs.path(f"/proc/net/{protocol}{version if version == 6 else ''}")

        contents = entry.open("rt")

        # Skip over the header row
        contents.readline()
        for line in contents:
            if not (line := line.strip()):
                continue

            socket = NetSocket.from_line(line, version)
            socket.protocol_string = entry.name

            if socket.protocol_string in ("udp", "raw", "udp6", "raw6"):
                socket.state_string = self.UDPStates(int(socket.state, 16)).name
            else:
                socket.state_string = self.TCPStates(int(socket.state, 16)).name

            user = self.target.user_details.find(uid=socket.uid)
            user_name = user.user.name if user else str(socket.uid)
            socket.owner = user_name

            # inode 0 could indicate a kernel process, which has no assiciated PID or FDs in /proc
            # or a inode of 0 could mean the socket is in a TIME_WAIT state.
            if socket.inode == 0:
                socket.pid = socket.inode
                yield socket
            else:
                processes = self.target.proc.inode_to_pids(socket.inode)

                for process in processes:
                    socket.pid = process.pid
                    socket.name = process.name
                    socket.cmdline = process.cmdline
                    yield socket

    def _parse_unix_sockets(self) -> Iterator[UnixSocket]:
        """Internal function to parse ``/proc/net/unix`` entries."""

        entry = self.target.fs.path("/proc/net/unix")
        contents = entry.open("rt")

        # Skip over the header row
        contents.readline()
        for line in contents:
            if not line.strip():
                continue

            socket = UnixSocket.from_line(line)

            socket.stream_type_string = self.SocketStreamType(socket.type).name
            socket.state_string = self.SocketStateType(socket.state).name

            yield socket

    def _parse_packet_sockets(self) -> Iterator[PacketSocket]:
        """Internal function to parse ``/proc/net/packet`` entries."""
        entry = self.target.fs.path("/proc/net/packet")
        contents = entry.open("rt")

        # Skip over the header row
        contents.readline()
        for line in contents:
            if not (line := line.strip()):
                continue

            socket = PacketSocket.from_line(line)

            processes = self.target.proc.inode_to_pids(socket.inode)

            socket.protocol_type = self.PacketProtocolTypes(socket.protocol).name

            user = self.target.user_details.find(uid=socket.user)
            user_name = user.user.name if user else str(socket.user)
            socket.owner = user_name

            for proc in processes:
                socket.pid = proc.pid
                socket.name = proc.name
                socket.cmdline = proc.cmdline
                yield socket


class ProcProcess:
    def __init__(self, target: Target, pid: int | str, proc_root: str = "/proc"):
        self.target = target
        self.root = proc_root
        self._pid = int(pid)

        # Note: ttys and pttys are not yet mapped to the process
        if self._pid == 0:
            # The process with PID 0 is responsible for paging and is referred to as the "swapper" or "sched" process.
            # It is a part of the kernel and is not a regular user-mode process.
            self.name = "swapper"
        else:
            self.entry = target.fs.path(fsutil.join(self.root, str(pid)))
            if not self.entry.exists():
                raise ProcessLookupError(f"Process with PID {pid} could not be found on target: {target}")

            self._stat_file = self._parse_proc_stat_entry()
            self.name = self._process_name

    def _parse_proc_status_entry(self) -> dict[str, str]:
        """Internal function to parse the contents of ``/proc/[pid]/status``."""
        status = self.get("status").open("rt")
        status_dict = {}

        for line in status.readlines():
            key, value = line.split(":", maxsplit=1)
            key = key.lower().strip()
            value = value.strip()
            status_dict[key] = value

        return status_dict

    def _parse_proc_stat_entry(self) -> dict[str, str | int]:
        """Internal function to parse the contents of ``/proc/[pid]/stat``."""
        status_dict = {}
        entry = self.get("stat")

        if entry.exists():
            status = entry.open("rt").readline()

            # The process name exists between parentheses in the second field.
            # Since the name can be arbitrary we have to find the first and last parentheses.
            start_name, end_name = status.find("("), status.rfind(")")

            head = status[:start_name]
            tail = status[end_name:]
            name = status[start_name + 1 : end_name]
            status = head + tail

            for idx, part in enumerate(status.split()[: len(PROC_STAT_NAMES)]):
                try:
                    part = int(part)
                except ValueError:
                    part = part

                status_dict[PROC_STAT_NAMES[idx]] = part

            status_dict["comm"] = name

        return status_dict

    def _parse_environ(self) -> Iterator[Environ]:
        """Internal function to parse entries in ``/proc/[pid]/environ``."""
        # entries in /proc/<pid>/environ are null-terminated
        if not (environ_path := self.get("environ")).exists():
            return

        for line in environ_path.read_text().split("\x00"):
            if line == "":
                # Skip empty line
                continue

            try:
                variable, contents = line.split("=", maxsplit=1)
            except ValueError:
                # Convention tells us that variable names and values are split on '='
                # in practice this is not always the case.
                variable = None
                contents = line

            yield Environ(variable, contents)

    @property
    def _boottime(self) -> int | None:
        """Returns the boot time of the system.

        Used internally to determine process start- and runtimes.
        """
        for line in self.target.fs.path(self.root).joinpath("stat").open("rt").readlines():
            if not line.startswith("btime"):
                continue

            return int(line.split()[1])
        return None

    def get(self, path: str) -> Path:
        """Returns a TargetPath relative to this process."""
        return self.entry.joinpath(path)

    @property
    def owner(self) -> str:
        """Return the username or the user ID (uid) (if owner is not found) of the owner of this process."""
        if self.uid:
            owner = self.target.user_details.find(uid=self.uid)
            return owner.user.name

        return str(self.uid)

    @property
    def uid(self) -> int:
        """Return the user ID (uid) of the owner of this process."""
        uid = int(self.get("loginuid").read_bytes())
        # loginuid can hold the value "4294967295" (0xFFFFFFFF).
        # Which is defined as "not set" and -1 should be returned.
        return -1 if uid == 0xFFFFFFFF else uid

    @property
    def pid(self) -> int:
        """Returns the process ID (pid) associated to this process."""
        return self._pid

    @property
    def parent(self) -> ProcProcess | None:
        """Returns the parent :class:`ProcProcess` of this process."""
        if self.pid == 0:
            return None
        return ProcProcess(self.target, self._stat_file.get("ppid"))

    @property
    def ppid(self) -> int | None:
        """Returns the parent process ID (ppid) associated to this process."""
        if parent := self.parent:
            return parent.pid
        return None

    @property
    def parent_name(self) -> str | None:
        """Returns the name associated to the parent process ID (ppid) of this process."""
        if parent := self.parent:
            return parent.name
        return None

    @property
    def state(self) -> str:
        """Returns the state of the process (S'leeping, R'unning, I'dle, etc)."""
        return ProcessStateEnum[self._stat_file.get("state", "N")].value

    @property
    def starttime(self) -> datetime | None:
        """Returns the start time of the process."""
        # Starttime is saved in clockticks per second from the boot time.
        # we asume a standard of 100 clockticks per second. the actual value can be obtained from `getconf CLK_TCK`
        starttime = self._stat_file.get("starttime", 0) / 100

        # Check if there is a boottime. If not, we cannot determine the starttime.
        if boottime := self._boottime:
            return from_unix(boottime + starttime)

        return None

    @property
    def runtime(self) -> timedelta | None:
        """Returns the runtime of a process until the moment of acquisition."""
        if starttime := self.starttime:
            return self.now - starttime

        return None

    @property
    def now(self) -> datetime:
        """Returns the ``now()`` timestamp of the system at the moment of acquisition."""
        return self.uptime + from_unix(self._boottime)

    def environ(self) -> Iterator[Environ]:
        """Yields the content of the environ file associated with the process."""
        yield from self._parse_environ()

    @property
    def uptime(self) -> timedelta:
        """Returns the uptime of the system from the moment it was acquired."""
        # uptime is saved in seconds from boottime
        uptime = self.target.fs.path(self.root).joinpath("uptime").read_text().split()[0]
        return timedelta(seconds=float(uptime))

    @property
    def _process_name(self) -> str:
        """Internal function that returns the name of the process."""
        return self._stat_file.get("comm", "")

    @property
    def cmdline(self) -> str:
        """Return the command line of a process."""

        line = ""
        entry = self.get("cmdline")

        if entry.exists():
            line = entry.read_text()
            # Cmdlines are null-terminated and use null-bytes to separate the different parts. Translate this back.
            line = line.rstrip("\x00").replace("\x00", " ")

        return line

    def stat(self) -> fsutil.stat_result:
        """Return a stat entry of the process."""
        return self.entry.stat()


class ProcPlugin(Plugin):
    __namespace__ = "proc"

    def __init__(self, target: Target):
        super().__init__(target)
        self.sockets = Sockets(self.target)

    def check_compatible(self) -> None:
        # The /proc folder can exist on live targets, but is empty.
        # So we make the check a little more specific using next()
        if not self.target.fs.exists("/proc") or not next(self.target.fs.iterdir("/proc"), False):
            raise UnsupportedPluginError("No /proc directory found")

    @cached_property
    def inode_map(self) -> dict[int, list[ProcProcess]]:
        """Creates a inode to pid mapping for all process IDs in ``/proc/[pid]``."""
        map = defaultdict(list)

        for path in self.iter_proc():
            if (fdpath := path.joinpath("fd")).exists():
                for fd in fdpath.iterdir():
                    link = fd.readlink().as_posix()

                    if link.startswith("socket:["):
                        _, inode = link.split(":")  # socket type, inode

                        # strip brackets from fd string (socket:[1337])
                        inode = int(inode[1:-1])
                        map[inode].append(ProcProcess(self.target, path.name))

        return map

    @internal
    def iter_proc(self) -> Iterator[Path]:
        """Yields ``/proc/[pid]`` filesystems entries for every process id (pid) found in procfs."""
        yield from self.target.fs.path("/proc").glob("[0-9]*")

    @internal
    def inode_to_pids(self, inode: int) -> list[ProcProcess]:
        return self.inode_map.get(inode, [])

    @internal
    def process(self, pid: int | str) -> ProcProcess:
        return ProcProcess(self.target, pid)

    @internal
    def processes(self) -> Iterator[ProcProcess]:
        for path in self.iter_proc():
            yield ProcProcess(self.target, path.name)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from dissect.util.ts import webkittimestamp
from flow.record.fieldtypes import datetime as dt

from dissect.target.helpers import keychain
from dissect.target.plugins.apps.browser.chrome import ChromePlugin
from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIPlugin
from tests._utils import absolute_path
from tests.conftest import add_win_user
from tests.plugins.os.windows.test__os import map_version_value
from tests.plugins.os.windows.test_lsa import (
    POLICY_KEY_PATH_NT6,
    map_lsa_polkey,
    map_lsa_secrets,
    map_lsa_system_keys,
)

if TYPE_CHECKING:
    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.helpers.regutil import VirtualHive
    from dissect.target.target import Target


@pytest.fixture
def target_chrome_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Default\\",
        absolute_path("_data/plugins/apps/browser/chrome/generic"),
    )
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Profile 1\\",
        absolute_path("_data/plugins/apps/browser/chrome/generic"),
    )

    target_win_users.add_plugin(ChromePlugin)

    return target_win_users


@pytest.fixture
def target_chrome_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Target:
    fs_unix.map_dir("/root/.config/google-chrome/Default/", absolute_path("_data/plugins/apps/browser/chrome/generic"))
    fs_unix.map_dir(
        "/root/.config/google-chrome/Profile 1/", absolute_path("_data/plugins/apps/browser/chrome/generic")
    )

    target_unix_users.add_plugin(ChromePlugin)

    return target_unix_users


@pytest.fixture
def target_chrome_win_snapshot(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Default",
        absolute_path("_data/plugins/apps/browser/chrome/generic"),
    )

    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Profile 1",
        absolute_path("_data/plugins/apps/browser/chrome/generic"),
    )

    target_win_users.add_plugin(ChromePlugin)

    return target_win_users


@pytest.mark.parametrize(
    "target_platform",
    ["target_chrome_win", "target_chrome_unix", "target_chrome_win_snapshot"],
)
def test_chrome_history(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chrome.history())

    assert len(records) == 10
    assert {"chrome"} == {record.browser for record in records}

    assert (
        records[0].url == "https://www.google.com/search?q=github+fox-it+dissect&oq=github+fox-it+dissect"
        "&aqs=chrome..69i57.12832j0j4&sourceid=chrome&ie=UTF-8"
    )
    assert records[0].id == 1
    assert records[0].visit_count == 2
    assert records[0].ts == dt("2023-02-24T11:54:07.157810+00:00")


@pytest.mark.parametrize(
    "target_platform",
    ["target_chrome_win", "target_chrome_unix", "target_chrome_win_snapshot"],
)
def test_chrome_downloads(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chrome.downloads())

    assert len(records) == 2
    assert {"chrome"} == {record.browser for record in records}

    assert records[0].id == 6
    assert records[0].ts_start == dt("2023-02-24T11:54:19.726147+00:00")
    assert records[0].ts_end == dt("2023-02-24T11:54:21.030043+00:00")
    assert records[0].url == "https://codeload.github.com/fox-it/dissect/zip/refs/heads/main"
    assert records[0].state == "complete"


@pytest.mark.parametrize(
    "target_platform",
    ["target_chrome_win", "target_chrome_unix", "target_chrome_win_snapshot"],
)
def test_chrome_extensions(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chrome.extensions())

    assert len(records) == 16
    assert {"chrome"} == {record.browser for record in records}

    assert records[0].ts_install == dt("2022-11-24T15:20:43.682152+00:00")
    assert records[0].ts_update == dt("2022-11-24T15:20:43.682152+00:00")
    assert records[0].name == "Web Store"
    assert records[0].version == "0.2"
    assert records[0].extension_id == "ahfgeienlihckogmohjhadlkjgocpleb"


def test_windows_chrome_passwords_plugin(target_chrome_win: Target) -> None:
    records = list(target_chrome_win.chrome.passwords())

    assert len(records) == 4

    for record in records:
        assert record.browser == "chrome"
        assert record.decrypted_username == "username"
        assert record.decrypted_password is None

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_chrome_passwords_basic_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.config/google-chrome/Default/", absolute_path("_data/plugins/apps/browser/chromium/unix/basic/")
    )
    fs_unix.map_dir(
        "/root/.config/google-chrome/Profile 1/", absolute_path("_data/plugins/apps/browser/chromium/unix/basic/")
    )
    target_unix_users.add_plugin(ChromePlugin)

    records = list(target_unix_users.chrome.passwords())

    assert len(records) == 4

    for record in records:
        assert record.browser == "chrome"
        assert record.decrypted_username == "username"
        assert record.decrypted_password == "password"

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_chrome_passwords_gnome_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.config/google-chrome/Default/", absolute_path("_data/plugins/apps/browser/chromium/unix/gnome/")
    )
    fs_unix.map_dir(
        "/root/.config/google-chrome/Profile 1/", absolute_path("_data/plugins/apps/browser/chromium/unix/gnome/")
    )
    target_unix_users.add_plugin(ChromePlugin)

    records = list(target_unix_users.chrome.passwords())

    assert len(records) == 2

    assert records[0].decrypted_username == "username"
    assert records[0].decrypted_password is None
    assert records[0].url == "https://test.com/"


@pytest.fixture
def target_win_users_dpapi(
    hive_hklm: VirtualHive, hive_hku: VirtualHive, fs_win: VirtualFilesystem, target_win: Target
) -> Target:
    # Add users
    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-1342509979-482553916-3960431919-1000",
        home="C:\\Users\\user",
    )

    # Add system dpapi files
    fs_win.map_dir(
        "Windows/System32/Microsoft/Protect",
        absolute_path("_data/plugins/os/windows/dpapi/fixture/windows_10/Protect_System32"),
    )

    # Add user dpapi files
    fs_win.map_dir(
        "Users/User/AppData/Roaming/Microsoft/Protect",
        absolute_path("_data/plugins/os/windows/dpapi/fixture/windows_10/Protect_User"),
    )

    # Add registry dpapi keys
    map_lsa_system_keys(hive_hklm, {"Data": "8fa8e1fb", "GBG": "a6e23eb8", "JD": "fe5ffdaf", "Skew1": "6e289261"})
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT6,
        bytes.fromhex(
            "00000001ecffe17b2a997440aa939adbff26f1fc0300000000000000676f5836"
            "37c3e0e7b9edf43b3b29b1d0d24cb6bfc60e0fc4dc446e7d244d303533b90a2b"
            "d732fcf985748a8917aea73e9d0b290ee4ba2f53e6a9a0ac9b3c9b26e721b01b"
            "7a6c1f92b517e2a33f5f6de7f736716793b19872059595e6b4dc888d19d1d615"
            "d602bed553478c411d2fed045602ddbb5adc31c9901021ad339bca368bdb554f"
            "fe074a7074528d5e9dcbb467"
        ),
    )
    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "000000017c3e71eca8fb4eed03ea4361fbc783870300000000000000af64ca32"
                "a15059f8e38f328a5f16d063939bdbb9321ba159dcafd9cdf316d82f89a829d7"
                "58024b276d099ef2290ca46fc7b2635568500bf2d31ed8ce1e0330345cca5ef3"
                "e8d18399a22ae88db12872ee5bb0c1f0dd3b83066269d0d9618b19bb"
            )
        },
    )

    target_win.add_plugin(DPAPIPlugin)
    return target_win


@pytest.fixture
def target_win_11_users_dpapi(
    hive_hklm: VirtualHive, hive_hku: VirtualHive, fs_win: VirtualFilesystem, target_win: Target
) -> Target:
    # Add users
    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="C:\\WINDOWS\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-3656658933-2463154391-3030686545-1001",
        home="C:\\Users\\User",
    )

    # Add system dpapi files
    fs_win.map_dir(
        "Windows/System32/Microsoft/Protect",
        absolute_path("_data/plugins/os/windows/dpapi/fixture/windows_11/Protect_System32"),
    )

    # Add user dpapi files
    fs_win.map_dir(
        "Users/User/AppData/Roaming/Microsoft/Protect",
        absolute_path("_data/plugins/os/windows/dpapi/fixture/windows_11/Protect_User"),
    )

    # Add registry dpapi keys
    map_lsa_system_keys(hive_hklm, {"Data": "5ef65665", "GBG": "df865f5a", "JD": "bdee0692", "Skew1": "73bc8e8c"})
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT6,
        bytes.fromhex(
            "00000001ecffe17b2a997440aa939adbff26f1fc0300000000000000da616d31"
            "4c9d457773de28cbf0ba721b98904cd68f7d0442ca7ad8e5409fe4d8bf7da10c"
            "8612389537bd7789d6bba9a4632c3ff90a91455dcea3869b87c04c3970b2f6f4"
            "8071b486b84c00d4fa3fab4f2f67578676fc4ef3072d9801b2ab062a758b4173"
            "8eee755b57dee2f59d42166c0827d7ecb33903b6a52eb6f5f96b6b9fba9a7ed0"
            "edf84a2be6b732cff0727e9a"
        ),
    )
    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "0000000195325072efa465ba92ec5edd44c4fefc030000000000000031291c7b"
                "4acb32b7c319c6fd70073435b0f4aa9f6e48451cd225382a0703b6505d75201d"
                "c5c10350492071c92bbc10bebbac687a6caed2d6f2f30d13ff5b744ad56727a8"
                "5bfb8ee9badcf2f784ca65228591356a7cff0aa5ecac645336d55389"
            )
        },
    )

    map_version_value(target_win, "ProductName", "Windows 10 Pro")
    map_version_value(target_win, "CurrentVersion", "10.0")
    map_version_value(target_win, "CurrentBuildNumber", "26100")

    target_win.add_plugin(DPAPIPlugin)
    return target_win


@pytest.mark.parametrize(
    ("keychain_value", "expected_password", "expected_notes"),
    [
        ("user", "StrongPassword", "Example note."),
        ("invalid", None, None),
    ],
)
def test_windows_chrome_passwords_dpapi(
    target_win_users_dpapi: Target,
    fs_win: VirtualFilesystem,
    keychain_value: str,
    expected_password: str | None,
    expected_notes: str | None,
) -> None:
    fs_win.map_dir(
        "Users/user/AppData/Local/Google/Chrome/User Data",
        absolute_path("_data/plugins/apps/browser/chrome/dpapi/windows_10/User_Data"),
    )

    map_version_value(target_win_users_dpapi, "CurrentVersion", 10.0)

    target_win_users_dpapi.add_plugin(ChromePlugin)

    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value=keychain_value,
        identifier=None,
        provider="user",
    )

    records = list(target_win_users_dpapi.chrome.passwords())

    assert len(keychain.get_all_keys()) == 1
    assert len(records) == 2

    assert records[0].url == "https://example.com/"
    assert records[0].encrypted_password == bytes.fromhex(
        "7631304fc7d5702f638993eb325f1075c1464a57cd02b4cf246ed0fd6cf8b2f1e9f5c4551b536a0b5fc973c411"
    )
    assert records[0].decrypted_password == expected_password
    assert records[0].encrypted_notes == bytes.fromhex(
        "76313052fa24300b1592880a5c3bdfbbcfab1fb10450a3bf385f7547cd76ec900a2ca37c5a6104dff0cbe404"
    )
    assert records[0].decrypted_notes == expected_notes


def test_windows_chrome_cookies_dpapi(target_win_users_dpapi: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_dir(
        "Users/user/AppData/Local/Google/Chrome/User Data",
        absolute_path("_data/plugins/apps/browser/chrome/dpapi/windows_10/User_Data"),
    )

    map_version_value(target_win_users_dpapi, "CurrentVersion", 10.0)

    target_win_users_dpapi.add_plugin(ChromePlugin)

    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="user",
        identifier=None,
        provider="user",
    )

    records = list(target_win_users_dpapi.chrome.cookies())

    assert len(records) == 4

    assert records[0].ts_created == webkittimestamp(13370000000000000)
    assert records[0].ts_last_accessed == webkittimestamp(13370000000000000)
    assert records[0].browser == "chrome"
    assert records[0].name == "tbb"
    assert records[0].value == "false"
    assert records[0].host == ".tweakers.net"
    assert records[0].is_secure

    assert {c.name: c.value for c in records} == {
        "tbb": "false",
        "twk-theme": "twk-light",
        "GPS": "1",
        "PREF": "tz=Europe.Berlin",
    }


def test_chrome_windows_snapshots(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    base_dirs = (
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Default",
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Profile 1",
    )
    snapshot_dirs = (
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Default",
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\119.0.7845.119\\Default",
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Profile 1",
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\119.0.7845.119\\Profile 1",
    )
    profile_dirs = base_dirs + snapshot_dirs

    for dir in profile_dirs:
        fs_win.map_dir(
            dir,
            absolute_path("_data/plugins/apps/browser/chrome/generic"),
        )

    target_win_users.add_plugin(ChromePlugin)

    records_list = [
        list(target_win_users.chrome.history()),
        list(target_win_users.chrome.extensions()),
        list(target_win_users.chrome.downloads()),
    ]

    # Loop over the different types of records and verify we have the same amount of records in each profile directory.
    for records in records_list:
        assert {"chrome"} == {record.browser for record in records}

        base_path_records = [r for r in records if str(r.source.parent).endswith(base_dirs)]

        # Retrieve records that are in the snapshot's directory.
        snapshot_records = [r for r in records if str(r.source.parent).endswith(snapshot_dirs)]

        # We map the same files in each of the snapshot directories.
        # We have two base directories and four snapshot directories, so we expect twice the amount of records.
        assert len(base_path_records) == len(snapshot_records) // 2


def test_chrome_windows_11_decryption(target_win_11_users_dpapi: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we can decrypt Windows 11 Google Chrome version 127/130 and newer passwords and cookies.

    Elevation Service usage by Chromium-based browsers (Google Chrome, Microsoft Edge) depend on several environment
    based circumstances (e.g. Windows version, feature flags, account log-in state). To force usage of the Elevation
    Service, run the command(s) below.

    .. code-block::

        (chrome.exe|msedge.exe) --enable-field-trial-config --enable-features=UseElevator,AppBoundEncryptionKeyV3

    """

    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="password",
        identifier=None,
        provider="user",
    )

    fs_win.map_dir(
        "Users/user/AppData/Local/Google/Chrome/User Data",
        absolute_path("_data/plugins/apps/browser/chrome/dpapi/windows_11/User_Data"),
    )

    target_win_11_users_dpapi.add_plugin(ChromePlugin)

    passwords = list(target_win_11_users_dpapi.chrome.passwords())
    assert len(passwords) == 2

    assert passwords[0].id == 1
    assert passwords[0].url == "https://elevated-example.com/"
    assert passwords[0].decrypted_username == "username@example.com"
    assert passwords[0].encrypted_password == bytes.fromhex(
        "763230b6ed2338175e5baa4daccc34697aa08809a69ead978a869cf11fabe0cafff7edf2340412"
    )
    assert passwords[0].decrypted_password == "password"
    assert passwords[0].encrypted_notes == bytes.fromhex(
        "763230892774a36593aefaa21416d3268235e98dd5cd7bc15cca023f669cd5df821066eba9de7bf0fe63f690cd"
    )
    assert passwords[0].decrypted_notes == "some note here"
    assert passwords[0].username == "User"

    assert passwords[1].id == 2
    assert passwords[1].url == "https://another-example.com/"
    assert passwords[1].decrypted_username == "username@domain.com"
    assert passwords[1].encrypted_password == bytes.fromhex(
        "763230e470c68414223312778d6345a548616a03be50106d6be9b0bd19b1a3186eecf7a41426a4eca7ed4eeaa0adf400e3c5"
    )
    assert passwords[1].decrypted_password == "MyPasswordIsSecret!"
    assert passwords[1].username == "User"

    cookies = list(target_win_11_users_dpapi.chrome.cookies())
    assert len(cookies) == 2

    assert cookies[0].host == "werkenvoornederland.nl"
    assert cookies[0].name == "ExampleCookieName"
    assert cookies[0].value == "1"

    assert cookies[1].host == "rijksoverheid.nl"
    assert cookies[1].name == "AnotherExampleCookieName"
    assert cookies[1].value == "420"


@pytest.mark.benchmark
def test_benchmark_chrome_userdirs(
    target_win_users: Target, fs_win: VirtualFilesystem, benchmark: BenchmarkFixture
) -> None:
    """Benchmark :class:`ChromiumMixin` ``_build_userdirs`` performance."""

    for i in range(3):
        fs_win.map_dir(
            f"Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Profile {i}",
            absolute_path("_data/plugins/apps/browser/chrome/generic"),
        )

    benchmark(
        lambda: target_win_users.add_plugin(ChromePlugin, check_compatible=True) and list(target_win_users.chrome())
    )



from __future__ import annotations

import textwrap
from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.webserver.nginx import NginxPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_nginx_txt(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log")
    fs_unix.map_file("var/log/nginx/access.log", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[0]
    assert record.ts == datetime(2022, 12, 1, 0, 3, 57, tzinfo=timezone.utc)
    assert record.status_code == 200
    assert record.remote_ip == "1.2.3.4"
    assert record.remote_user == "admin"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_nginx_ipv6(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log")
    fs_unix.map_file("var/log/nginx/access.log", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[1]
    assert record.remote_ip == "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    assert record.status_code == 200
    assert record.remote_user == "-"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_nginx_gz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log.gz")
    fs_unix.map_file("var/log/nginx/access.log.1.gz", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[0]
    assert record.status_code == 200
    assert record.remote_ip == "1.2.3.4"
    assert record.remote_user == "admin"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_nginx_bz2(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log.bz2")
    fs_unix.map_file("var/log/nginx/access.log.1.bz2", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[0]
    assert record.status_code == 200
    assert record.remote_ip == "1.2.3.4"
    assert record.remote_user == "admin"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_nginx_config(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config_file = absolute_path("_data/plugins/apps/webserver/nginx/nginx.conf")
    fs_unix.map_file("etc/nginx/nginx.conf", config_file)

    for i, log in enumerate(["access.log", "domain1.access.log", "domain2.access.log", "big.server.access.log"]):
        fs_unix.map_file_fh(f"opt/logs/{i}/{log}", BytesIO(b"Foo"))

    target_unix.add_plugin(NginxPlugin)

    assert len(target_unix.nginx.access_paths) == 4
    assert len(target_unix.nginx.error_paths) == 0


def test_nginx_config_logs_logrotated(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config_file = absolute_path("_data/plugins/apps/webserver/nginx/nginx.conf")
    fs_unix.map_file("etc/nginx/nginx.conf", config_file)
    fs_unix.map_file_fh("opt/logs/0/access.log", BytesIO(b"Foo1"))
    fs_unix.map_file_fh("opt/logs/0/access.log.1", BytesIO(b"Foo2"))
    fs_unix.map_file_fh("opt/logs/0/access.log.2", BytesIO(b"Foo3"))
    fs_unix.map_file_fh("opt/logs/1/domain1.access.log", BytesIO(b"Foo4"))
    fs_unix.map_file_fh("var/log/nginx/access.log", BytesIO(b"Foo5"))

    target_unix.add_plugin(NginxPlugin)

    assert len(target_unix.nginx.access_paths) == 5
    assert len(target_unix.nginx.error_paths) == 0


def test_nginx_config_commented_logs(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config = """
    # access_log      /foo/bar/old.log main;
    access_log      /foo/bar/new.log main;

    # error_log         /foo/bar/error/old.log warn;
    error_log               /foo/bar/error/new.log;
    """
    fs_unix.map_file_fh("etc/nginx/nginx.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("foo/bar/new.log", BytesIO(b"New"))
    fs_unix.map_file_fh("foo/bar/old.log", BytesIO(b"Old"))
    fs_unix.map_file_fh("foo/bar/error/new.log", BytesIO())
    fs_unix.map_file_fh("foo/bar/error/old.log", BytesIO())

    target_unix.add_plugin(NginxPlugin)

    assert len(target_unix.nginx.access_paths) == 2
    assert len(target_unix.nginx.error_paths) == 2

    assert sorted(map(str, target_unix.nginx.access_paths)) == ["/foo/bar/new.log", "/foo/bar/old.log"]
    assert sorted(map(str, target_unix.nginx.error_paths)) == ["/foo/bar/error/new.log", "/foo/bar/error/old.log"]


def test_nginx_error_logs(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we detect and parse nginx error logs correctly."""

    errors = """
    2025/01/31 13:37:01 [alert] 12345#12345: this is a message
    2025/01/31 13:37:02 [alert] 12345#12345: this is another message
    2025/01/31 13:37:03 [alert] 12345#12345: and a third message!
    """
    fs_unix.map_file_fh("var/log/nginx/error.log", BytesIO(textwrap.dedent(errors).encode()))

    target_unix.add_plugin(NginxPlugin)
    records = list(target_unix.nginx.error())

    assert len(records) == 3

    assert records[0].ts == datetime(2025, 1, 31, 13, 37, 1, tzinfo=timezone.utc)
    assert records[0].level == "alert"
    assert records[0].message == "this is a message"
    assert records[0].source == "/var/log/nginx/error.log"


def test_nginx_parse_config(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we parse config files and their include directives correctly."""

    base_conf = """
    user www www;
    server {
        listen 1337;
        server_name example;
        index index.html;
        root /var/www/html;
        include some.conf;
        access_log /some/access.log;
    }
    include /more/confs/*.conf;
    """
    fs_unix.map_file_fh("/etc/nginx/nginx.conf", BytesIO(textwrap.dedent(base_conf).encode()))

    some_conf = """
    error_log /some/error.log;
    """
    fs_unix.map_file_fh("/etc/nginx/some.conf", BytesIO(textwrap.dedent(some_conf).encode()))

    more_confs_one = """
    server {
        listen 80;
        server_name eighty;
        index index.html;
        root /var/www/eighty;
        access_log /eighty/access.log;
        include /bla/foo.conf;
    }
    """
    fs_unix.map_file_fh("/more/confs/one.conf", BytesIO(textwrap.dedent(more_confs_one).encode()))

    foo_conf = """
    error_log /eighty/error.log;
    """
    fs_unix.map_file_fh("/bla/foo.conf", BytesIO(textwrap.dedent(foo_conf).encode()))

    fs_unix.map_file_fh("/some/access.log", BytesIO())
    fs_unix.map_file_fh("/some/error.log", BytesIO())
    fs_unix.map_file_fh("/eighty/access.log.1", BytesIO())
    fs_unix.map_file_fh("/eighty/error.log.1", BytesIO())

    target_unix.add_plugin(NginxPlugin)

    assert sorted(map(str, target_unix.nginx.access_paths)) == [
        "/eighty/access.log.1",
        "/some/access.log",
    ]
    assert sorted(map(str, target_unix.nginx.error_paths)) == [
        "/eighty/error.log.1",
        "/some/error.log",
    ]

    assert sorted(map(str, target_unix.nginx.host_paths)) == [
        "/etc/nginx/nginx.conf",
        "/more/confs/one.conf",
    ]

    records = sorted(target_unix.nginx.hosts(), key=lambda r: r.source)

    assert len(records) == 2

    assert records[0].ts
    assert records[0].server_name == "example"
    assert records[0].server_port == 1337
    assert records[0].root_path == "/var/www/html"
    assert records[0].access_log_config == "/some/access.log"
    assert not records[0].error_log_config
    assert records[0].source == "/etc/nginx/nginx.conf"

    assert records[1].ts
    assert records[1].server_name == "eighty"
    assert records[1].server_port == 80
    assert records[1].root_path == "/var/www/eighty"
    assert records[1].access_log_config == "/eighty/access.log"
    assert not records[1].error_log_config
    assert records[1].source == "/more/confs/one.conf"


def test_nginx_host_certificate(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can parse NGINX host TLS certificates."""

    bar_conf = """
        server {
        listen              443 ssl;
        ssl_certificate     /path/to/cert.pem;
        ssl_certificate_key /path/to/key.pem;
        server_name         example.com;
        root                /path/to/www;
    }
    """
    fs_unix.map_file_fh("/etc/nginx/sites-enabled/example.com.conf", BytesIO(textwrap.dedent(bar_conf).encode()))
    fs_unix.map_file("/path/to/cert.pem", absolute_path("_data/plugins/apps/webserver/certificates/example.crt"))
    fs_unix.map_file("/path/to/key.pem", absolute_path("_data/plugins/apps/webserver/certificates/example.key"))
    target_unix.add_plugin(NginxPlugin)

    records = list(target_unix.nginx.hosts())
    assert len(records) == 1
    assert records[0].server_name == "example.com"
    assert records[0].server_port == 443
    assert records[0].tls_certificate == "/path/to/cert.pem"

    records = list(target_unix.nginx.certificates())
    assert len(records) == 1

    # x509
    assert records[0].fingerprint.md5 == "a218ac9b6dbdaa8b23658c4d18c1cfc1"
    assert records[0].fingerprint.sha1 == "6566d8ebea1feb4eb3d12d9486cddb69e4e9e827"
    assert records[0].fingerprint.sha256 == "7221d881743505f13b7bfe854bdf800d7f0cd22d34307ed7157808a295299471"
    assert records[0].serial_number == 21067204948278457910649605551283467908287726794
    assert records[0].serial_number_hex == "03b0afa702c33e37fffd40e0c402b2120c1284ca"
    assert records[0].issuer_dn == "C=AU,ST=Some-State,O=Internet Widgits Pty Ltd,CN=example.com"



from __future__ import annotations

from enum import IntEnum
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.filesystem import FilesystemEntry
    from dissect.target.target import Target

CapabilityRecord = TargetRecordDescriptor(
    "filesystem/unix/capability",
    [
        ("datetime", "mtime"),
        ("path", "path"),
        ("string[]", "permitted"),
        ("string[]", "inheritable"),
        ("boolean", "effective"),
        ("uint32", "root_id"),
    ],
)


VFS_CAP_REVISION_MASK = 0xFF000000
VFS_CAP_REVISION_SHIFT = 24
VFS_CAP_FLAGS_MASK = (~VFS_CAP_REVISION_MASK) & 0xFFFFFFFF
VFS_CAP_FLAGS_EFFECTIVE = 0x000001

VFS_CAP_REVISION_1 = 0x01000000
VFS_CAP_U32_1 = 1

VFS_CAP_REVISION_2 = 0x02000000
VFS_CAP_U32_2 = 2

VFS_CAP_REVISION_3 = 0x03000000
VFS_CAP_U32_3 = 2


class Capabilities(IntEnum):
    CAP_CHOWN = 0
    CAP_DAC_OVERRIDE = 1
    CAP_DAC_READ_SEARCH = 2
    CAP_FOWNER = 3
    CAP_FSETID = 4
    CAP_KILL = 5
    CAP_SETGID = 6
    CAP_SETUID = 7
    CAP_SETPCAP = 8
    CAP_LINUX_IMMUTABLE = 9
    CAP_NET_BIND_SERVICE = 10
    CAP_NET_BROADCAST = 11
    CAP_NET_ADMIN = 12
    CAP_NET_RAW = 13
    CAP_IPC_LOCK = 14
    CAP_IPC_OWNER = 15
    CAP_SYS_MODULE = 16
    CAP_SYS_RAWIO = 17
    CAP_SYS_CHROOT = 18
    CAP_SYS_PTRACE = 19
    CAP_SYS_PACCT = 20
    CAP_SYS_ADMIN = 21
    CAP_SYS_BOOT = 22
    CAP_SYS_NICE = 23
    CAP_SYS_RESOURCE = 24
    CAP_SYS_TIME = 25
    CAP_SYS_TTY_CONFIG = 26
    CAP_MKNOD = 27
    CAP_LEASE = 28
    CAP_AUDIT_WRITE = 29
    CAP_AUDIT_CONTROL = 30
    CAP_SETFCAP = 31
    CAP_MAC_OVERRIDE = 32
    CAP_MAC_ADMIN = 33
    CAP_SYSLOG = 34
    CAP_WAKE_ALARM = 35
    CAP_BLOCK_SUSPEND = 36
    CAP_AUDIT_READ = 37
    CAP_PERFMON = 38
    CAP_BPF = 39
    CAP_CHECKPOINT_RESTORE = 40


class CapabilityPlugin(Plugin):
    """Plugin to yield files with capabilites set."""

    def check_compatible(self) -> None:
        if not self.target.has_function("walkfs"):
            raise UnsupportedPluginError("Need walkfs plugin")

        if not any(fs.__type__ in ("extfs", "xfs") for fs in self.target.filesystems):
            raise UnsupportedPluginError("Capability plugin only works on EXT and XFS filesystems")

    @export(record=CapabilityRecord)
    def capability_binaries(self) -> Iterator[CapabilityRecord]:
        """Find all files that have capabilities set on files.

        References:
            - https://github.com/torvalds/linux/blob/master/include/uapi/linux/capability.h
        """

        for entry in self.target.fs.recurse("/"):
            if not entry.is_file(follow_symlinks=False):
                continue
            yield from parse_entry(entry, self.target)


def parse_entry(entry: FilesystemEntry, target: Target) -> Iterator[CapabilityRecord]:
    try:
        attrs = [attr for attr in entry.lattr() if attr.name == "security.capability"]
    except Exception as e:
        target.log.warning("Failed to get attrs for entry %s", entry)
        target.log.debug("", exc_info=e)
        return

    for attr in attrs:
        try:
            permitted, inheritable, effective, root_id = parse_attr(attr.value)
        except ValueError as e:
            target.log.warning("Could not parse attributes for entry %s: %s", entry, str(e.value))
            target.log.debug("", exc_info=e)

        yield CapabilityRecord(
            mtime=entry.lstat().st_mtime,
            path=target.fs.path(entry.path),
            permitted=permitted,
            inheritable=inheritable,
            effective=effective,
            root_id=root_id,
            _target=target,
        )


def parse_attr(attr: bytes) -> tuple[list[str], list[str], bool, int]:
    """Efficiently parse a Linux xattr capability struct.

    Returns:
        A tuple of permitted capability names, inheritable capability names, effective flag and ``root_id``.
    """
    buf = BytesIO(attr)

    # The struct is small enough we can just use int.from_bytes
    magic_etc = int.from_bytes(buf.read(4), "little")
    effective = magic_etc & VFS_CAP_FLAGS_EFFECTIVE != 0
    cap_revision = magic_etc & VFS_CAP_REVISION_MASK

    permitted_caps = []
    inheritable_caps = []
    root_id = None

    if cap_revision == VFS_CAP_REVISION_1:
        num_caps = VFS_CAP_U32_1
        data_len = (1 + 2 * VFS_CAP_U32_1) * 4

    elif cap_revision == VFS_CAP_REVISION_2:
        num_caps = VFS_CAP_U32_2
        data_len = (1 + 2 * VFS_CAP_U32_2) * 4

    elif cap_revision == VFS_CAP_REVISION_3:
        num_caps = VFS_CAP_U32_3
        data_len = (2 + 2 * VFS_CAP_U32_2) * 4

    else:
        raise ValueError(f"Unexpected capability revision '{cap_revision}'")

    if data_len != (actual_len := len(attr)):
        raise ValueError("Unexpected capability length (%s vs %s)", data_len, actual_len)

    for _ in range(num_caps):
        permitted_caps.append(int.from_bytes(buf.read(4), "little"))
        inheritable_caps.append(int.from_bytes(buf.read(4), "little"))

    if cap_revision == VFS_CAP_REVISION_3:
        root_id = int.from_bytes(buf.read(4), "little")

    permitted = []
    inheritable = []

    for capability in Capabilities:
        for caps, results in [(permitted_caps, permitted), (inheritable_caps, inheritable)]:
            # CAP_TO_INDEX
            cap_index = capability.value >> 5
            if cap_index >= len(caps):
                # We loop over all capabilities, but might only have a version 1 caps list
                continue

            if caps[cap_index] & (1 << (capability.value & 31)) != 0:
                results.append(capability.name)

    return permitted, inheritable, effective, root_id



from __future__ import annotations

import math
from typing import TYPE_CHECKING, BinaryIO

import dissect.apfs as apfs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry, VirtualFilesystem
from dissect.target.helpers import fsutil, keychain
from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Iterator
    from uuid import UUID

    from dissect.apfs.objects.fs import DirectoryEntry, INode

log = get_logger(__name__)


class ApfsFilesystem(Filesystem):
    __type__ = "apfs-container"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.container = apfs.APFS(fh)
        self.vfs = VirtualFilesystem()
        self._volumes = []

        for vol in self.container.volumes:
            try:
                vol = ApfsVolumeFilesystem(self, uuid=vol.uuid)
            except FilesystemError as e:  # noqa: PERF203
                log.error("Failed to open APFS volume %r (%s): %s", vol.name, vol.uuid, str(e))  # noqa: TRY400
                log.debug("", exc_info=e)
            else:
                self._volumes.append(vol)
                self.vfs.map_file_entry(vol.volume.name, vol.get("/"))

    def __repr__(self) -> str:
        return f"<Filesystem type={self.__type__} volumes={len(self._volumes)}>"

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        block = fh.read(4096)
        return block[32:36] == b"NXSB"

    def iter_subfs(self) -> Iterator[ApfsVolumeFilesystem]:
        yield from self._volumes

    def get(self, path: str) -> FilesystemEntry:
        return self.vfs.get(path)


class ApfsVolumeFilesystem(Filesystem):
    __type__ = "apfs"

    def __init__(self, container: ApfsFilesystem, name: str | None = None, uuid: UUID | None = None):
        if name is not None and uuid is not None:
            raise ValueError("Only one of name or uuid is allowed")

        self.container = container

        if name:
            self.apfs = next((vol for vol in self.container.container.volumes if vol.name == name), None)
        elif uuid:
            self.apfs = next((vol for vol in self.container.container.volumes if vol.uuid == uuid), None)
        else:
            raise ValueError("Either name or uuid must be provided")

        if not self.apfs:
            raise ValueError("No APFS volume found with the specified identifier")

        if self.apfs.is_encrypted:
            kek_uuids = [str(kek.uuid) for kek in self.apfs.keybag.keks()]
            for key in keychain.get_keys_for_provider(self.__type__) + keychain.get_keys_without_provider():
                # Allow unlocking with keys that have no identifier or match the volume UUID or KEK UUIDs
                if key.identifier and (
                    key.identifier.lower() != str(self.apfs.uuid).lower() or key.identifier not in kek_uuids
                ):
                    continue

                try:
                    # But only pass the key identifier if it's a KEK UUID
                    self.apfs.unlock(key.value, key.identifier if key.identifier in kek_uuids else None)
                    break
                except apfs.Error:
                    continue
            else:
                raise FilesystemError("No valid decryption key found")

        super().__init__(
            container.volume,
            alt_separator=container.alt_separator,
            case_sensitive=not self.apfs.is_case_insensitive,
        )

    def __repr__(self) -> str:
        return f"<Filesystem type={self.__type__} name={self.apfs.name} uuid={self.apfs.uuid}>"

    def get(self, path: str) -> FilesystemEntry:
        return ApfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: INode | None = None) -> INode:
        try:
            return self.apfs.get(path, node)
        except apfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except apfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except apfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except apfs.Error as e:
            raise FileNotFoundError(path) from e


class ApfsDirEntry(DirEntry):
    fs: ApfsVolumeFilesystem
    entry: DirectoryEntry

    def get(self) -> ApfsFilesystemEntry:
        return ApfsFilesystemEntry(self.fs, self.path, self.entry.inode)

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return super().is_dir(follow_symlinks=follow_symlinks)

        return self.entry.is_dir()

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return super().is_file(follow_symlinks=follow_symlinks)

        return self.entry.is_file()

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class ApfsFilesystemEntry(FilesystemEntry):
    fs: ApfsVolumeFilesystem
    entry: INode

    def get(self, path: str) -> FilesystemEntry:
        entry_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        entry = self.fs._get_node(path, self.entry)
        return ApfsFilesystemEntry(self.fs, entry_path, entry)

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[FilesystemEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry in self._resolve().entry.iterdir():
            if entry.name in (".", ".."):
                continue

            yield ApfsDirEntry(self.fs, self.path, entry.name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError(self.path)

        return self.entry.readlink()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        entry = self.entry

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                entry.mode,
                entry.oid,
                0,
                entry.inode.nlink,
                entry.uid,
                entry.gid,
                entry.size,
                # timestamp() returns a float which will fill both the integer and float fields
                entry.atime.timestamp(),
                entry.mtime.timestamp(),
                entry.ctime.timestamp(),
            ]
        )

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = entry.inode.access_time
        st_info.st_mtime_ns = entry.inode.mod_time
        st_info.st_ctime_ns = entry.inode.change_time

        st_info.st_birthtime = entry.btime.timestamp()
        st_info.st_birthtime_ns = entry.inode.create_time

        # Add block information of the filesystem
        st_info.st_blksize = self.fs.container.container.block_size

        st_info.st_blocks = 0
        if not self.is_dir():
            st_info.st_blocks = (st_info.st_blksize // 512) * math.ceil(st_info.st_size / st_info.st_blksize)

        return st_info



from __future__ import annotations

from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.linux.debian.snap import SnapPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_snap_packages(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if snap packages are discovered on unix systems."""

    fs_unix.map_file_fh("/etc/hostname", BytesIO(b"hostname"))
    fs_unix.map_file(
        "/var/lib/snapd/snaps/firefox_12345.snap",
        absolute_path("_data/plugins/os/unix/linux/debian/snap/firefox.snap"),
    )
    fs_unix.map_file(
        "/var/lib/snapd/snaps/firefox_67890.snap",
        absolute_path("_data/plugins/os/unix/linux/debian/snap/firefox.snap"),
    )

    target_unix_users.add_plugin(UnixPlugin)
    target_unix_users.add_plugin(SnapPlugin)

    assert target_unix_users.has_function("snap")

    results = list(target_unix_users.snaps())
    assert len(results) == 2

    assert results[0].hostname == "hostname"
    assert results[0].ts_modified == datetime(2024, 9, 17, 13, 18, 58, tzinfo=timezone.utc)
    assert results[0].name == "firefox"
    assert results[0].version == "129.0.2-1"
    assert results[0].author is None
    assert results[0].type is None
    assert results[0].path == "/var/lib/snapd/snaps/firefox_12345.snap"



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from dissect.target.filesystem import LayerFilesystem
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers.logging import get_logger
from dissect.target.loaders.tar import TarSubLoader

if TYPE_CHECKING:
    import tarfile
    from pathlib import Path

    from dissect.target.target import Target


log = get_logger(__name__)

DOCKER_ARCHIVE_IMAGE = {
    "manifest.json",
    "repositories",
}

OCI_IMAGE = {
    "blobs",
    "oci-layout",
    "index.json",
}


class ContainerImageTarSubLoader(TarSubLoader):
    """Load saved container images.

    Supports both the Docker and OCI image specifications.

    Tested with output from ``docker image save`` and ``podman image save``.

    References:
        - https://snyk.io/blog/container-image-formats/
        - https://github.com/moby/docker-image-spec/
        - https://github.com/opencontainers/image-spec/
    """

    def __init__(self, path: Path, tar: tarfile.TarFile, **kwargs):
        super().__init__(path, tar, **kwargs)

        self.tarfs: TarFilesystem = None
        self.layers: list[Path] = []

        self.manifest = None
        self.name = None
        self.config = None

        try:
            self.tarfs = TarFilesystem(None, tarfile=tar)
        except Exception as e:
            raise ValueError(f"Unable to open {tar} as TarFilesystem: {e}") from e

        # Moby/Docker spec uses manifest.json
        if self.tarfs.path("/manifest.json").exists():
            try:
                self.manifest = json.loads(self.tarfs.path("/manifest.json").read_text())[0]
                self.name = (self.manifest.get("RepoTags") or [None])[0]
                self.layers = [self.tarfs.path(p) for p in self.manifest.get("Layers", [])]
            except Exception as e:
                raise ValueError(f"Unable to read manifest.json inside docker image filesystem: {e}") from e

            try:
                self.config = json.loads(self.tarfs.path(self.manifest.get("Config")).read_text())
            except Exception as e:
                raise ValueError(f"Unable to read config inside docker image filesystem: {e}") from e

            # When self.name not set from "RepoTags" in manifest.json, try creating one from the "Config" json.
            if self.config and not self.name:
                labels = self.config.get("config", {}).get("Labels", {})
                title = labels.get("org.opencontainers.image.title")
                version = labels.get("org.opencontainers.image.version")
                if title and version:
                    self.name = f"{title}:{version}"

        # OCI spec only has index.json
        elif self.tarfs.path("/index.json").exists():
            try:
                index = json.loads(self.tarfs.path("/index.json").read_text())
                self.config = json.loads(
                    self.tarfs.path("/blobs").joinpath(index["manifests"][0]["digest"].replace(":", "/")).read_text()
                )
                self.layers = [
                    self.tarfs.path("/blobs").joinpath(layer["digest"].replace(":", "/"))
                    for layer in self.config.get("layers", [])
                ]
            except Exception as e:
                raise ValueError(f"Unable to load OCI container: {e}") from e

    @staticmethod
    def detect(path: Path, tar: tarfile.TarFile) -> bool:
        names = tar.getnames()
        return OCI_IMAGE.issubset(names) or DOCKER_ARCHIVE_IMAGE.issubset(names)

    def map(self, target: Target) -> None:
        fs = LayerFilesystem()

        for layer in self.layers:
            if not layer.exists():
                log.warning("Layer %s does not exist in container image", layer)
                continue

            fs.append_fs_layer(TarFilesystem(layer.open("rb")))

        fs.append_layer().mount("$fs$/container", self.tarfs)

        target.filesystems.add(fs)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.database.sqlite3 import SQLite3
from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.target.target import Target

ActivitiesCacheRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/activitiescache",
    [
        ("datetime", "start_time"),
        ("datetime", "end_time"),
        ("datetime", "last_modified_time"),
        ("datetime", "last_modified_on_client"),
        ("datetime", "original_last_modified_on_client"),
        ("datetime", "expiration_time"),
        ("string", "activity_id"),
        ("string", "app_id"),
        ("string", "enterprise_id"),
        ("string", "app_activity_id"),
        ("string", "group_app_activity_id"),
        ("string", "group"),
        ("uint32", "activity_type"),
        ("uint32", "activity_status"),
        ("uint32", "activity_priority"),
        ("string", "match_id"),
        ("uint32", "etag"),
        ("string", "tag"),
        ("boolean", "is_local_only"),
        ("datetime", "created_in_cloud"),
        ("string", "platform_device_id"),
        ("string", "package_id_hash"),
        ("string", "payload"),
        ("string", "original_payload"),
        ("string", "clipboard_payload"),
        ("path", "source"),
    ],
)


class ActivitiesCachePlugin(Plugin):
    """Plugin that parses the ActivitiesCache.db on newer Windows 10 machines.

    References:
        - https://www.cclsolutionsgroup.com/resources/technical-papers
        - https://salt4n6.com/2018/05/03/windows-10-timeline-forensic-artefacts/
    """

    def __init__(self, target: Target):
        super().__init__(target)
        self.cachefiles = []

        for user_details in target.user_details.all_with_home():
            full_path = user_details.home_path.joinpath("AppData/Local/ConnectedDevicesPlatform")
            cache_files = full_path.glob("*/ActivitiesCache.db")
            for cache_file in cache_files:
                if cache_file.exists():
                    self.cachefiles.append((user_details.user, cache_file))

    def check_compatible(self) -> None:
        if len(self.cachefiles) == 0:
            raise UnsupportedPluginError("No ActiviesCache.db files found")

    @export(record=ActivitiesCacheRecord)
    def activitiescache(self) -> Iterator[ActivitiesCacheRecord]:
        """Return ActivitiesCache.db database content.

        The Windows Activities Cache database keeps track of activity on a device, such as application and services
        usage, files opened, and websites browsed. This database file can therefore be used to create a system timeline.
        It has first been used on Windows 10 1803.

        Currently only puts the database records straight into Flow Records. Ideally
        we do some additional parsing on this later.

        References:
            - https://artifacts-kb.readthedocs.io/en/latest/sources/windows/ActivitiesCacheDatabase.html
            - https://salt4n6.com/2018/05/03/windows-10-timeline-forensic-artefacts/

        Yields ActivitiesCacheRecords with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            start_time (datetime): StartTime field.
            end_time (datetime): EndTime field.
            last_modified_time (datetime): LastModifiedTime field.
            last_modified_on_client (datetime): LastModifiedOnClient field.
            original_last_modified_on_client (datetime): OriginalLastModifiedOnClient field.
            expiration_time (datetime): ExpirationTime field.
            activity_id (string): Id field in hex.
            app_id (string): AppId field, JSON string containing multiple types of app name definitions.
            enterprise_id (string): EnterpriseId field.
            app_activity_id (string): AppActivityId field.
            group_app_activity_id (string): GroupAppActivityId field.
            group (string): Group field.
            activity_type (int): ActivityType field.
            activity_status (int): ActivityStatus field.
            activity_priority (int): Priority field.
            match_id (string): MatchId field.
            etag (int): ETag field.
            tag (string): Tag field.
            is_local_only (boolean): IsLocalOnly field.
            created_in_cloud (datetime): CreatedInCloud field.
            platform_device_id (string): PlatformDeviceId field.
            package_id_hash (string): PackageIdHash field.
            payload (string): Payload field. JSON string containing payload data, varies per type.
            original_payload (string): OriginalPayload field.
            clipboard_payload (string): ClipboardPayload field.
        """
        for user, cache_file in self.cachefiles:
            with SQLite3(cache_file) as db:
                if table := db.table("Activity"):
                    for r in table.rows():
                        yield ActivitiesCacheRecord(
                            start_time=mkts(r["[StartTime]"]),
                            end_time=mkts(r["[EndTime]"]),
                            last_modified_time=mkts(r["[LastModifiedTime]"]),
                            last_modified_on_client=mkts(r["[LastModifiedOnClient]"]),
                            original_last_modified_on_client=mkts(r["[OriginalLastModifiedOnClient]"]),
                            expiration_time=mkts(r["[ExpirationTime]"]),
                            activity_id=r["[Id]"].hex(),
                            app_id=r["[AppId]"],
                            enterprise_id=r["[EnterpriseId]"] or None,
                            app_activity_id=r["[AppActivityId]"],
                            group_app_activity_id=r["[GroupAppActivityId]"] or None,
                            group=r["[Group]"],
                            activity_type=r["[ActivityType]"],
                            activity_status=r["[ActivityStatus]"],
                            activity_priority=r["[Priority]"],
                            match_id=r["[MatchId]"],
                            etag=r["[ETag]"],
                            tag=r["[Tag]"],
                            is_local_only=r["[IsLocalOnly]"],
                            created_in_cloud=r["[CreatedInCloud]"],
                            platform_device_id=r["[PlatformDeviceId]"],
                            package_id_hash=r["[PackageIdHash]"],
                            payload=r["[Payload]"],
                            original_payload=r["[OriginalPayload]"],
                            clipboard_payload=r["[ClipboardPayload]"],
                            source=cache_file,
                            _target=self.target,
                            _user=user,
                        )


def mkts(ts: int) -> datetime | None:
    """Timestamps inside ActivitiesCache.db are stored in a Unix-like format.

    Source: https://salt4n6.com/2018/05/03/windows-10-timeline-forensic-artefacts/
    """
    return from_unix(ts) if ts else None



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugin import export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class FreeBsdPlugin(BsdPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self._os_release = self._parse_os_release("/bin/freebsd-version*")

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        FREEBSD_PATHS = {
            "/.sujournal",
            "/entropy",
            "/bin/freebsd-version",
        }

        for fs in target.filesystems:
            if fs.exists("/net") and any(fs.exists(path) for path in FREEBSD_PATHS):
                return fs
        return None

    @export(property=True)
    def version(self) -> str | None:
        return self._os_release.get("USERLAND_VERSION")



from __future__ import annotations

import importlib
from pathlib import Path

TYPES = []
MAP = set()
PATTERNS: set[tuple[int, tuple[str]]] = set()

files = [
    Path(__file__).parent.joinpath("freedesktop.py").resolve(),
    *[p for p in Path(__file__).parent.glob("*.py") if p.stem not in ("__init__", "freedesktop", "overrides")],
    Path(__file__).parent.joinpath("overrides.py").resolve(),
]

for file in files:
    obj = importlib.import_module(f"dissect.target.helpers.magic.mimetypes.{file.stem}")

    if not hasattr(obj, "TYPES"):
        raise ValueError(f"Module {obj!s} does not have a TYPES attribute.")

    if not isinstance(obj.TYPES, list):
        raise TypeError(f"Attribute {obj!s}.TYPES is not a list.")

    TYPES += obj.TYPES

for i, definition in enumerate(TYPES):
    if "type" not in definition:
        raise ValueError(f"Magic definition missing type: {definition}")

    if "name" not in definition:
        raise ValueError(f"Magic definition missing name: {definition}")

    if "magic" not in definition and "pattern" not in definition:
        raise ValueError(f"Magic definition missing magic or pattern: {definition}")

    if "pattern" in definition:
        # We only support patterns like `*.foo`, not `foo.*` or `foo*bar`
        PATTERNS.add((i, tuple([p.replace("*", "") for p in definition.get("pattern", [])])))

    for magic in definition.get("magic", []):
        if "value" not in magic:
            raise ValueError(f"Magic definition missing value in {definition}")

        MAP.add((i, magic.get("offset", 0), magic["value"]))



from __future__ import annotations

import re
from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

PATTERN_IPTABLES_SAVE_GENERATED = re.compile(r"# Generated by (ip6?tables-save) v([.\d]+) on (.+)")

PATTERN_IPTABLES_SAVE_POLICY = re.compile(
    r":(?P<chain>\S+) (?P<policy>[\w-]+) \[(?P<packet_count>\d+):(?P<byte_count>\d+)]"
)

PATTERN_IPTABLES_SAVE_RULE = re.compile(
    r"(?:\[(?P<packet_count>\d+):(?P<byte_count>\d+)] )?(?P<rule>-A (?P<chain>\S+).+)"
)

IptablesSaveRecord = TargetRecordDescriptor(
    "linux/iptables/save",
    [
        ("datetime", "ts"),
        ("string", "program"),
        ("string", "version"),
        ("string", "table"),
        ("string", "chain"),
        ("string", "type"),
        ("string", "rule"),
        ("varint", "packet_count"),
        ("varint", "byte_count"),
        ("path", "source"),
    ],
)


class IptablesSavePlugin(Plugin):
    """Parser for iptables-save, ip6tables-save and ufw rules.

    As iptables rules are not stored on disk by default, users
    that want persistent rules need to store them somewhere and
    reload them on boot. iptables provides tools to save and reload
    rules (iptables-save and iptables-restore). These tools
    do not have a default path to look for rules, however there
    are multiple commonly used paths.

    References:
        - https://git.netfilter.org/iptables/
        - https://manpages.ubuntu.com/manpages/jammy/en/man8/ufw-framework.8.html
    """

    COMMON_SAVE_PATHS = (
        # IPv4
        "/etc/iptables/rules.v4",
        "/etc/sysconfig/iptables",
        "/etc/iptables.rules",
        "/etc/iptablesRule.v4",
        "/etc/network/iptables.rules",
        # IPv6
        "/etc/iptables/rules.v6",
        "/etc/iptablesRule.v6",
        "/etc/sysconfig/ip6tables",
        # UFW
        "/etc/ufw/before.rules",
        "/etc/ufw/user.rules",
        "/etc/ufw/after.rules",
        "/etc/ufw/before6.rules",
        "/etc/ufw/user6.rules",
        "/etc/ufw/after6.rules",
        "/usr/share/ufw/user.rules",
        "/usr/share/ufw/user6.rules",
        "/usr/share/ufw/iptables/user.rules",
        "/usr/share/ufw/iptables/user6.rules",
        "/var/lib/ufw/user.rules",
        "/var/lib/ufw/user6.rules",
    )

    LOG_TIME_FORMAT = "%a %b  %d %H:%M:%S %Y"

    def __init__(self, target: Target):
        super().__init__(target)
        self._rule_files = list(self._get_rule_files())

    def check_compatible(self) -> None:
        if not len(self._rule_files):
            raise UnsupportedPluginError("No iptables rules found")

    def _get_rule_files(self) -> Iterator[Path]:
        """Yield the paths of iptables-save output files."""
        for rule_path in self.COMMON_SAVE_PATHS:
            rule_path = self.target.fs.path(rule_path)

            if rule_path.exists():
                with rule_path.open("r") as h_rule:
                    first_line = h_rule.readline()

                if PATTERN_IPTABLES_SAVE_GENERATED.match(first_line) or "*filter" in first_line:
                    yield rule_path

    @export(record=IptablesSaveRecord)
    def iptables(self) -> Iterator[IptablesSaveRecord]:
        """Return iptables and ufw rules saved using iptables-save."""
        target_tz = self.target.datetime.tzinfo

        for rule_path in self._rule_files:
            current_program = None
            current_version = None
            current_ts = None
            current_table = None

            with rule_path.open("r") as h_rule:
                for line in h_rule:
                    line = line.strip()

                    if not line:
                        continue

                    elif match := PATTERN_IPTABLES_SAVE_GENERATED.match(line):
                        current_program, current_version, ts_string = match.groups()
                        current_ts = datetime.strptime(ts_string, self.LOG_TIME_FORMAT).replace(
                            tzinfo=target_tz,
                        )

                    elif line.startswith("#") or line == "COMMIT":
                        continue

                    # Table
                    elif line.startswith("*"):
                        current_table = line.removeprefix("*")

                    # Policy
                    elif match := PATTERN_IPTABLES_SAVE_POLICY.match(line):
                        policy = match.groupdict()
                        yield IptablesSaveRecord(
                            ts=current_ts,
                            program=current_program,
                            version=current_version,
                            table=current_table,
                            chain=policy["chain"],
                            type="policy",
                            rule=policy["policy"],
                            packet_count=policy["packet_count"],
                            byte_count=policy["byte_count"],
                            source=rule_path,
                        )

                    # Rule
                    elif match := PATTERN_IPTABLES_SAVE_RULE.match(line):
                        rule = match.groupdict()
                        yield IptablesSaveRecord(
                            ts=current_ts,
                            program=current_program,
                            version=current_version,
                            table=current_table,
                            chain=rule["chain"],
                            type="rule",
                            rule=rule["rule"],
                            packet_count=rule["packet_count"] or None,
                            byte_count=rule["byte_count"] or None,
                            source=rule_path,
                        )

                    else:
                        self.target.log.warning("Unable to parse line: %s in file %s", line, rule_path)



from __future__ import annotations

import re
from dataclasses import dataclass, field
from datetime import datetime, timezone
from ipaddress import IPv4Interface, IPv6Interface, ip_address, ip_interface
from itertools import chain, islice
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, get_args

from dissect.target.exceptions import PluginError
from dissect.target.helpers import configutil
from dissect.target.helpers.record import UnixInterfaceRecord
from dissect.target.helpers.utils import to_list
from dissect.target.plugin import arg, export
from dissect.target.plugins.os.default.network import InterfaceRecord, NetworkPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from ipaddress import IPv4Address, IPv6Address

    from dissect.target.plugins.os.unix.log.journal import JournalRecord
    from dissect.target.plugins.os.unix.log.messages import MessagesRecord
    from dissect.target.target import Target, TargetPath

    NetAddress = IPv4Address | IPv6Address
    NetInterface = IPv4Interface | IPv6Interface
    LogRecord = JournalRecord | MessagesRecord


class LinuxNetworkPlugin(NetworkPlugin):
    """Linux network interface plugin."""

    @arg(
        "-s",
        "--syslog",
        action="store_true",
        help="scan syslog for dhcp leases; use --max-lines to specify number of lines",
    )
    @arg(
        "-m",
        "--max-lines",
        action="store",
        type=int,
        help="maximum number of lines of syslog to scan (implies --syslog); default is 1000 if --syslog is set",
    )
    @export(record=get_args(InterfaceRecord))
    def interfaces(self, syslog: bool = False, max_lines: int | None = None) -> Iterator[UnixInterfaceRecord]:
        """Yield interfaces."""
        # If max_lines is specified, syslog is assumed True
        if max_lines is not None:
            syslog = True
        # If syslog is True and max_lines is None, use default 1000
        if syslog and max_lines is None:
            max_lines = 1000
        max_lines = None if max_lines == 0 else max_lines
        yield from super().interfaces(scan_syslog=syslog, max_lines=max_lines)

    def _interfaces(self, scan_syslog: bool, max_lines: int | None) -> Iterator[UnixInterfaceRecord]:
        """Try all available network configuration managers and aggregate the results."""
        for manager_cls in MANAGERS + LEASERS:
            manager: LinuxNetworkConfigParser = manager_cls(self.target)
            yield from manager.interfaces()

        if scan_syslog:
            syslog_parser = SyslogConfigParser(self.target, max_lines)
            yield from syslog_parser.interfaces()

    @export(record=UnixInterfaceRecord)
    def dhcp(self) -> Iterator[UnixInterfaceRecord]:
        """Return interfaces obtained via DHCP."""
        for interface in self.interfaces():
            if interface.type == "dhcp":
                yield interface


VlanIdByInterface = dict[str, set[int]]


class LinuxNetworkConfigParser:
    def __init__(self, target: Target):
        self._target = target

    def _config_files(self, config_paths: list[str], glob: str) -> list[TargetPath]:
        """Returns all configuration files in config_paths matching the given extension."""
        all_files = []
        for config_path in config_paths:
            paths = self._target.fs.path(config_path).glob(glob)
            all_files.extend(config_file for config_file in paths if config_file.is_file())

        return sorted(all_files, key=lambda p: p.stem)

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        """Parse network interfaces from configuration files."""
        yield from ()


class NetworkManagerConfigParser(LinuxNetworkConfigParser):
    """NetworkManager configuration parser.

    NetworkManager configuration files are generally in an INI-like format.
    Note that Red Hat and Fedora deprecated ifcfg files.
    Documentation: https://networkmanager.dev/docs/api/latest/nm-settings-keyfile.html
    """

    config_paths: tuple[str, ...] = (
        "/etc/NetworkManager/system-connections/",
        "/usr/lib/NetworkManager/system-connections/",
        "/run/NetworkManager/system-connections/",
    )

    @dataclass
    class ParserContext:
        source: str
        uuid: str | None = None
        last_connected: datetime | None = None
        name: str | None = None
        mac_address: str | None = None
        type: str = ""
        dns: set[NetAddress] = field(default_factory=set)
        ip_interfaces: set[NetInterface] = field(default_factory=set)
        gateways: set[NetAddress] = field(default_factory=set)
        dhcp_ipv4: bool = False
        dhcp_ipv6: bool = False
        vlan: set[int] = field(default_factory=set)

        def to_record(self, target: Target) -> UnixInterfaceRecord:
            return UnixInterfaceRecord(
                name=self.name,
                type=self.type,
                enabled=None,
                cidr=self.ip_interfaces,
                gateway=list(self.gateways),
                dns=list(self.dns),
                mac=to_list(self.mac_address),
                source=self.source,
                dhcp_ipv4=self.dhcp_ipv4,
                dhcp_ipv6=self.dhcp_ipv6,
                last_connected=self.last_connected,
                vlan=list(self.vlan),
                configurator="NetworkManager",
                _target=target,
            )

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        connections: list[NetworkManagerConfigParser.ParserContext] = []
        vlan_id_by_interface: VlanIdByInterface = {}

        for connection_file_path in self._config_files(self.config_paths, "*"):
            try:
                config = configutil.parse(connection_file_path, hint="ini")
                context = self.ParserContext(source=connection_file_path)
                common_section: dict[str, str] = config.get("connection", {})
                context.type = common_section.get("type", "")
                sub_type: dict[str, str] = config.get(context.type, {})

                if context.type == "vlan":
                    self._parse_vlan(sub_type, vlan_id_by_interface)
                    continue

                for ip_version in ["ipv4", "ipv6"]:
                    ip_section: dict[str, str] = config.get(ip_version, {})
                    for key, value in ip_section.items():
                        self._parse_ip_section_key(key, value, context, ip_version)

                context.name = common_section.get("interface-name")
                context.mac_address = sub_type.get("mac-address")
                context.uuid = common_section.get("uuid")
                context.source = str(connection_file_path)
                context.last_connected = self._parse_lastconnected(common_section.get("timestamp", ""))

                connections.append(context)

            except Exception as e:
                self._target.log.warning("Error parsing network config file %s", connection_file_path)
                self._target.log.debug("", exc_info=e)

        for connection in connections:
            vlan_ids_from_interface = vlan_id_by_interface.get(connection.name, set())
            connection.vlan.update(vlan_ids_from_interface)

            vlan_ids_from_uuid = vlan_id_by_interface.get(connection.uuid, set())
            connection.vlan.update(vlan_ids_from_uuid)

            yield connection.to_record(self._target)

    def _parse_route(self, route: str) -> NetAddress | None:
        """Parse a route and return gateway IP address."""
        if (elements := route.split(",")) and len(elements) > 1:
            return ip_address(elements[1])

        return None

    def _parse_lastconnected(self, last_connected: str) -> datetime | None:
        """Parse last connected timestamp."""
        if not last_connected:
            return None

        return datetime.fromtimestamp(int(last_connected), timezone.utc)

    def _parse_ip_section_key(
        self,
        key: str,
        value: str,
        context: ParserContext,
        ip_version: Literal["ipv4", "ipv6"],
    ) -> None:
        if not (trimmed := value.strip()):
            return

        if key == "dns":
            context.dns.update(ip_address(addr) for addr in trimmed.split(";") if addr)
        elif key.startswith("address"):
            # Undocumented: single gateway on address line. Observed when running:
            # nmcli connection add type ethernet ... ip4 192.168.2.138/24 gw4 192.168.2.1
            ip, *gateway = trimmed.split(",", 1)
            context.ip_interfaces.add(ip_interface(ip))
            if gateway:
                context.gateways.add(ip_address(gateway[0]))
        elif key.startswith("gateway"):
            context.gateways.add(ip_address(trimmed))
        elif key == "method":
            if ip_version == "ipv4":
                context.dhcp_ipv4 = trimmed == "auto"
            elif ip_version == "ipv6":
                context.dhcp_ipv6 = trimmed == "auto"
        elif key.startswith("route") and (gateway := self._parse_route(value)):
            context.gateways.add(gateway)

    def _parse_vlan(self, sub_type: dict[str, Any], vlan_id_by_interface: VlanIdByInterface) -> None:
        parent_interface = sub_type.get("parent")
        vlan_id = sub_type.get("id")
        if not parent_interface or not vlan_id:
            return

        ids = vlan_id_by_interface.setdefault(parent_interface, set())
        ids.add(int(vlan_id))


class SystemdNetworkConfigParser(LinuxNetworkConfigParser):
    """Systemd network configuration parser.

    Systemd network configuration files are generally in an INI-like format with some quirks.
    Note that drop-in directories are not yet supported.

    Documentation: https://www.freedesktop.org/software/systemd/man/latest/systemd.network.html
    """

    collapsable_items: tuple[str, ...] = (
        "Match",
        "Network",
        "Link",
        "MACAddress",
        "Name",
        "Type",
    )

    config_paths: tuple[str, ...] = (
        "/etc/systemd/network/",
        "/run/systemd/network/",
        "/usr/lib/systemd/network/",
        "/usr/local/lib/systemd/network/",
    )

    class DhcpConfig(NamedTuple):
        ipv4: bool
        ipv6: bool

    # Can be enclosed in brackets for IPv6. Can also have port, iface name, and SNI, which we ignore.
    # Example: [1111:2222::3333]:9953%ifname#example.com
    dns_ip_patttern = re.compile(
        r"(?P<withoutBrackets>(?:\d{1,3}\.){3}\d{1,3})|\[(?P<withBrackets>\[?[0-9a-fA-F:]+\]?)\]"
    )

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        virtual_networks = self._parse_virtual_networks()
        yield from self._parse_networks(virtual_networks)

    def _parse_virtual_networks(self) -> VlanIdByInterface:
        """Parse virtual network configurations from systemd network configuration files."""

        virtual_networks: VlanIdByInterface = {}
        for config_file in self._config_files(self.config_paths, "*.netdev"):
            try:
                virtual_network_config = configutil.parse(config_file, hint="systemd")
                net_dev_section: dict[str, str] = virtual_network_config.get("NetDev", {})
                if net_dev_section.get("Kind") != "vlan":
                    continue

                vlan_id = virtual_network_config.get("VLAN", {}).get("Id")
                if (name := net_dev_section.get("Name")) and vlan_id:
                    vlan_ids = virtual_networks.setdefault(name, set())
                    vlan_ids.add(int(vlan_id))
            except Exception as e:
                self._target.log.warning("Error parsing virtual network config file %s", config_file)
                self._target.log.debug("", exc_info=e)

        return virtual_networks

    def _parse_networks(self, virtual_networks: VlanIdByInterface) -> Iterator[UnixInterfaceRecord]:
        """Parse network configurations from systemd network configuration files."""
        for config_file in self._config_files(self.config_paths, "*.network"):
            try:
                config = configutil.parse(config_file, hint="systemd", collapse=self.collapsable_items)

                match_section: dict[str, str] = config.get("Match", {})
                network_section: dict[str, str] = config.get("Network", {})
                link_section: dict[str, str] = config.get("Link", {})

                ip_interfaces: set[NetInterface] = set()
                gateways: set[NetAddress] = set()
                dns: set[NetAddress] = set()
                mac_addresses: set[str] = set()

                if link_mac := link_section.get("MACAddress"):
                    mac_addresses.add(link_mac)
                mac_addresses.update(match_section.get("MACAddress", "").split())
                mac_addresses.update(match_section.get("PermanentMACAddress", "").split())

                dns_value = to_list(network_section.get("DNS", []))
                dns.update(self._parse_dns_ip(dns_ip) for dns_ip in dns_value)

                address_value = to_list(network_section.get("Address", []))
                ip_interfaces.update(ip_interface(addr) for addr in address_value)

                gateway_value = to_list(network_section.get("Gateway", []))
                gateways.update(ip_address(gateway) for gateway in gateway_value)

                vlan_ids: set[int] = set()
                vlan_names = to_list(network_section.get("VLAN", []))
                for vlan_name in vlan_names:
                    if ids := virtual_networks.get(vlan_name):
                        vlan_ids.update(ids)

                route_sections: list[dict[str, Any]] = config.get("Route", [])
                gateway_values = (to_list(route_section.get("Gateway", [])) for route_section in route_sections)
                gateways.update(
                    filter(
                        None,
                        map(self._parse_gateway, chain.from_iterable(gateway_values)),
                    )
                )

                dhcp_ipv4, dhcp_ipv6 = self._parse_dhcp(network_section.get("DHCP"))

                yield UnixInterfaceRecord(
                    name=match_section.get("Name"),
                    type=match_section.get("Type"),
                    enabled=None,  # Unknown, dependent on run-time state
                    cidr=ip_interfaces,
                    gateway=list(gateways),
                    dns=list(dns),
                    mac=list(mac_addresses),
                    source=str(config_file),
                    dhcp_ipv4=dhcp_ipv4,
                    dhcp_ipv6=dhcp_ipv6,
                    vlan=list(vlan_ids),
                    configurator="systemd-networkd",
                    _target=self._target,
                )
            except Exception as e:  # noqa: PERF203
                self._target.log.warning("Error parsing network config file %s", config_file)
                self._target.log.debug("", exc_info=e)

    def _parse_dns_ip(self, address: str) -> NetAddress:
        """Parse DNS address from systemd network configuration file.

        The optional brackets and port number make this hard to parse.
        See https://www.freedesktop.org/software/systemd/man/latest/systemd.network.html and search for DNS.
        """

        if match := self.dns_ip_patttern.search(address):
            return ip_address(match.group("withoutBrackets") or match.group("withBrackets"))

        raise ValueError(f"Invalid DNS address format: {address}")

    def _parse_dhcp(self, value: str | None) -> DhcpConfig:
        """Parse DHCP value from systemd network configuration file to a named tuple (ipv4, ipv6)."""

        if value is None or value == "no":
            return self.DhcpConfig(ipv4=False, ipv6=False)
        if value == "yes":
            return self.DhcpConfig(ipv4=True, ipv6=True)
        if value == "ipv4":
            return self.DhcpConfig(ipv4=True, ipv6=False)
        if value == "ipv6":
            return self.DhcpConfig(ipv4=False, ipv6=True)

        raise ValueError(f"Invalid DHCP value: {value}")

    def _parse_gateway(self, value: str | None) -> NetAddress | None:
        if (not value) or (value in {"_dhcp4", "_ipv6ra"}):
            return None

        return ip_address(value)


class DhclientLeaseParser(LinuxNetworkConfigParser):
    """Parse network interfaces from dhclient DHCP ``.leases`` files.

    References:
        - https://linux.die.net/man/5/dhclient.conf
    """

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        """Parse network interfaces from dhclient DHCP ``.leases`` files."""

        # TODO: Add support for renew, rebind, expire events.
        for file in self._config_files(["/var/lib/dhclient/", "/var/lib/dhcp/"], "*.leases"):
            try:
                leases = configutil.parse(file, hint="leases")
                for label, data in leases.parsed_data.items():
                    if "lease" in label:
                        if (fixed_address := data.get("fixed-address")) is not None:
                            if (subnet_mask := data.get("option", {}).get("subnet-mask")) is not None:
                                fixed_address = ip_interface((fixed_address, subnet_mask))
                            else:
                                fixed_address = ip_interface(fixed_address)

                        if (router := data.get("option", {}).get("routers")) is not None:
                            router = ip_address(router)

                        yield UnixInterfaceRecord(
                            name=data.get("interface").strip('"'),
                            type="dhcp",
                            cidr={fixed_address},
                            gateway={router},
                            source=file,
                            dhcp_ipv4=isinstance(fixed_address, IPv4Interface),
                            dhcp_ipv6=isinstance(fixed_address, IPv6Interface),
                            configurator="dhclient",
                            _target=self._target,
                        )
            except Exception as e:  # noqa: PERF203
                self._target.log.warning("Error parsing dhclient leases file %s", file)
                self._target.log.debug("", exc_info=e)
                continue


class NetworkManagerLeaseParser(LinuxNetworkConfigParser):
    """Parse network interfaces from NetworkManager DHCP ``.lease`` files."""

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        """Parse network interfaces from NetworkManager DHCP ``.lease`` files."""
        for file in self._config_files(["/var/lib/NetworkManager/"], "*.lease"):
            try:
                lease = configutil.parse(file, hint="env")

                interface = file.stem.split("-")[-1]  # Extract interface name from file stem
                if ip_address := lease.get("ADDRESS"):
                    ip_address = ip_interface(ip_address)

                yield UnixInterfaceRecord(
                    name=interface,
                    type="dhcp",
                    cidr={ip_address},
                    source=file,
                    configurator="NetworkManager",
                    dhcp_ipv4=isinstance(ip_address, IPv4Interface),
                    dhcp_ipv6=isinstance(ip_address, IPv6Interface),
                    _target=self._target,
                )
            except Exception as e:  # noqa: PERF203
                self._target.log.warning("Error parsing NetworkManager lease file %s", file)
                self._target.log.debug("", exc_info=e)
                continue


class ProcConfigParser(LinuxNetworkConfigParser):
    """Parser for dynamic network configuration data from /proc/net.

    Parse gateways, interface names and network from /proc/net/route.
    Corroborate with TCP connections from /proc/net/tcp to find local IP addresses.
    Locally bound Ipv6 addresses are parsed from /proc/net/if_inet6.
    """

    # Regex to match lines like: |-- 127.0.0.1
    trie_ip_line_re = re.compile(r"^\s*\|\-\-\s*(\d+\.\d+\.\d+\.\d+)\s*$")

    @dataclass
    class ParserContext:
        name: str | None = None
        ip_interfaces: set[NetInterface] = field(default_factory=set)
        gateways: set[NetAddress] = field(default_factory=set)

        def to_record(self, target: Target) -> UnixInterfaceRecord:
            return UnixInterfaceRecord(
                name=self.name,
                cidr=self.ip_interfaces,
                gateway=list(self.gateways),
                source="/proc/net",
                configurator="proc",
                _target=target,
            )

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        (routes, interfaces) = self._parse_proc_net_route()

        tcp_local_ipv4 = self._parse_proc_net_tcp_local_ipv4()
        fib_tree_ipv4 = self._parse_proc_fib_trie()
        self._correlate_route_addresses(routes, tcp_local_ipv4 | fib_tree_ipv4, interfaces)

        self._parse_proc_net_if_inet6(interfaces)

        for iface in interfaces.values():
            yield iface.to_record(self._target)

    def _parse_proc_net_route(
        self,
    ) -> tuple[dict[str, set[IPv4Interface]], dict[str, ProcConfigParser.ParserContext]]:
        try:
            with self._target.fs.path("/proc/net/route").open("r") as f:
                lines = f.readlines()
        except FileNotFoundError:
            self._target.log.info("File /proc/net/route not found")
            return ({}, {})
        except Exception as e:
            self._target.log.warning("Error reading /proc/net/route")
            self._target.log.debug("", exc_info=e)
            return ({}, {})

        interfaces: dict[str, ProcConfigParser.ParserContext] = {}
        route_interfaces: dict[str, set[IPv4Interface]] = {}
        for line in lines[1:]:  # Skip header
            fields = line.split()
            if len(fields) < 8:
                self._target.log.warning("Skipping malformed line in /proc/net/route: %s", line)
                continue

            iface_name, destination_hex, gateway_hex, *_, mask = fields[:8]

            # Only add CIDR if not default route
            if (addr := be_hex_to_int(destination_hex)) != 0:
                mask_bit_count = be_hex_to_int(mask).bit_count()
                route_interfaces.setdefault(iface_name, set()).add(ip_interface((addr, mask_bit_count)))

            # Add gateway if not 0.0.0.0
            if (gateway := be_hex_to_int(gateway_hex)) != 0:
                iface = interfaces.setdefault(iface_name, ProcConfigParser.ParserContext(name=iface_name))
                iface.gateways.add(ip_address(gateway))

        return (route_interfaces, interfaces)

    def _correlate_route_addresses(
        self,
        routes: dict[str, set[IPv4Interface]],
        locally_bound_addresses: set[IPv4Address],
        interfaces: dict[str, ProcConfigParser.ParserContext],
    ) -> None:
        """Correlate routes with local bound addresses."""

        for iface_name, route_ifaces in routes.items():
            iface = interfaces.setdefault(iface_name, ProcConfigParser.ParserContext(name=iface_name))
            for route_iface in route_ifaces:
                matched_route_iface = False
                for ipv4 in locally_bound_addresses:
                    if ipv4 in route_iface.network:
                        iface.ip_interfaces.add(ip_interface((ipv4, route_iface.network.prefixlen)))
                        matched_route_iface = True
                        break
                if not matched_route_iface:
                    iface.ip_interfaces.add(route_iface)  # If no local IP found, still add the route interface

    def _parse_proc_net_if_inet6(self, ctx: dict[str, ProcConfigParser.ParserContext]) -> None:
        # Parse IPv6 from /proc/net/if_inet6
        try:
            with self._target.fs.path("/proc/net/if_inet6").open("r") as f:
                for line in f:
                    parts = line.split()
                    if len(parts) < 6:
                        self._target.log.warning("Skipping malformed line in /proc/net/if_inet6: %s", line)
                        continue
                    ipv6_hex, _, prefixlen_hex, _, _, iface_name = parts
                    iface = ctx.setdefault(iface_name, ProcConfigParser.ParserContext(name=iface_name))
                    ipv6_addr = int(ipv6_hex, 16)
                    prefix_len = int(prefixlen_hex, 16)
                    iface.ip_interfaces.add(ip_interface((ipv6_addr, prefix_len)))

        except FileNotFoundError:
            self._target.log.info("File /proc/net/if_inet6 does not exist")
        except Exception as e:
            self._target.log.warning("Error parsing /proc/net/if_inet6")
            self._target.log.debug("", exc_info=e)

    def _parse_proc_net_tcp_local_ipv4(self) -> set[IPv4Address]:
        try:
            with self._target.fs.path("/proc/net/tcp").open("r") as f:
                lines = f.readlines()
        except FileNotFoundError:
            self._target.log.info("File /proc/net/tcp does not exist")
            return set()
        except Exception as e:
            self._target.log.warning("Error reading /proc/net/tcp")
            self._target.log.debug("", exc_info=e)
            return set()

        local_ips = set()
        for line in lines[1:]:  # skip header
            fields = line.split()
            if len(fields) < 4:
                continue
            _, local_address, _, state, *_ = fields
            if state == "0A":  # 0A: systemd pollutes outgoing connections so we filter them out
                continue

            try:
                ip_hex, _ = local_address.split(":", 1)
                local_ips.add(ip_address(be_hex_to_int(ip_hex)))
            except Exception:
                self._target.log.warning("Failed to parse local address in /proc/net/tcp: %s", local_address)
                continue
        return local_ips

    def _parse_proc_fib_trie(self) -> set[IPv4Address]:
        """
        Parse a fib_trie-like text and return a set of IPv4Address objects
        for addresses ending in '/32 host LOCAL'.
        """

        try:
            with self._target.fs.path("/proc/net/fib_trie").open("r") as f:
                lines = f.readlines()
        except FileNotFoundError:
            self._target.log.info("File /proc/net/fib_trie does not exist")
            return set()
        except Exception as e:
            self._target.log.warning("Error reading /proc/net/fib_trie")
            self._target.log.debug("", exc_info=e)
            return set()

        result = set()
        iterator = configutil.PeekableIterator(lines)
        for line in iterator:
            # Skip all the lines that do not contain '/32 host LOCAL' as the next line
            next_line = iterator.peek()
            if not next_line or "/32 host LOCAL" not in next_line:
                continue

            if ip_match := self.trie_ip_line_re.match(line):
                result.add(ip_address(ip_match.group(1)))

        return result


class SyslogConfigParser(LinuxNetworkConfigParser):
    def __init__(self, target: Target, max_lines: int | None = None):
        super().__init__(target)
        self._max_lines = max_lines

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        parsers = [
            parse_ubuntu_cloud_init_dhcp_lease,
            parse_network_manager_dhcp_lease_old,
            parse_network_manager_dhcp_lease_new,
            parse_networkd_dhcp_lease,
            parse_network_manager_centos_dhcp_lease,
            parse_debian_centos_dhclient_lease,
        ]

        for record in self._read_logs():
            for parser in parsers:
                if dhcp_lease := parser(record):
                    yield UnixInterfaceRecord(
                        name=dhcp_lease.name,
                        cidr={dhcp_lease.interface},
                        gateway={dhcp_lease.gateway} if dhcp_lease.gateway else set(),
                        source="syslog",
                    )
                    break

    def _read_logs(self) -> Iterator[LogRecord]:
        try:
            messages_iter = islice(self._target.messages(), self._max_lines)
        except PluginError:
            self._target.log.debug("Could not search for DHCP leases from the messages plugin")
            messages_iter = iter(())

        try:
            journal_iter = islice(self._target.journal(), self._max_lines)
        except PluginError:
            self._target.log.debug("Could not search for DHCP leases from the journal plugin")
            journal_iter = iter(())

        return filter(lambda log: log.message, chain(messages_iter, journal_iter))


def be_hex_to_int(be_hex: str) -> int:
    """Convert big-endian hex string to integer."""
    return int.from_bytes(bytes.fromhex(be_hex), "little")


class DhcpLease(NamedTuple):
    name: str | None
    interface: NetInterface
    gateway: NetAddress | None


def parse_ubuntu_cloud_init_dhcp_lease(log_record: LogRecord) -> DhcpLease | None:
    """Parse DHCP lease information from Ubuntu cloud-init logs."""

    if not (match := re.search(r"Received dhcp lease on (\w*) for (\S+)", log_record.message)):
        return None
    name, interface = match.groups()

    return DhcpLease(name, ip_interface(interface), None) if interface else None


def parse_networkd_dhcp_lease(log_record: LogRecord) -> DhcpLease | None:
    """Parse DHCP lease information from systemd-networkd."""

    if not (match := re.search(r"(\S+): DHCPv[4|6] address (\S+) via (\S+)", log_record.message)):
        return None

    name, interface, gateway = match.groups()
    gateway_address = ip_address(gateway) if gateway else None
    return DhcpLease(name, ip_interface(interface), gateway_address) if interface else None


def parse_network_manager_dhcp_lease_old(log_record: LogRecord) -> DhcpLease | None:
    """Parse DHCP lease information from NetworkManager logs old style."""

    # dhcp4 (eth0): option ip_address           => '10.13.37.1'
    if not (match := re.search(r"dhcp[46] \((\S+)\): option ip_address\s+=>\s+'(\S+)'", log_record.message)):
        return None
    name, interface = match.groups()

    return DhcpLease(name, ip_interface(interface), None) if interface else None


def parse_network_manager_dhcp_lease_new(log_record: LogRecord) -> DhcpLease | None:
    """Parse DHCP lease information from NetworkManager logs new style."""

    # dhcp4 (eth0): state changed new lease, address=10.13.37.1
    if not (
        match := re.search(
            r"dhcp[46] \((\S+)\): state changed new lease, address=([^\s,]+)",
            log_record.message,
        )
    ):
        return None
    name, interface = match.groups()

    return DhcpLease(name, ip_interface(interface), None) if interface else None


def parse_network_manager_centos_dhcp_lease(log_record: LogRecord) -> DhcpLease | None:
    """Parse DHCP centos information"""

    if not (match := re.search(r"dhcp[46] \((\S+)\):\s+address\s+(\S+)", log_record.message)):
        return None
    name, interface = match.groups()
    return DhcpLease(name, ip_interface(interface), None) if interface else None


def parse_debian_centos_dhclient_lease(log_record: LogRecord) -> DhcpLease | None:
    """Parse DHCP lease information from dhclient bound log lines."""

    if getattr(log_record, "service", None) != "dhclient":
        return None

    # Apr  4 13:37:04 localhost dhclient[4]: bound to 10.13.37.4 -- renewal in 1337 seconds.
    if not (match := re.search(r"bound to (\S+)", log_record.message)):
        return None
    return DhcpLease(None, ip_interface(ip), None) if (ip := match.group(1)) else None


MANAGERS = [NetworkManagerConfigParser, SystemdNetworkConfigParser, ProcConfigParser]
LEASERS = [DhclientLeaseParser, NetworkManagerLeaseParser]



from __future__ import annotations

import json as jsonlib
from typing import TYPE_CHECKING, Any

from dissect.database.sqlite3 import SQLite3

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, internal

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class ConfigstorePlugin(Plugin):
    """Plugin to interact with the ESXi configstore."""

    __namespace__ = "configstore"

    def __init__(self, target: Target):
        super().__init__(target)
        self._configstore = {}

        # ESXi 7 introduced the configstore
        # It's made available at /etc/vmware/configstore/current-store-1 during boot, but stored at
        # the path used below in local.tgz
        if (path := target.fs.path("/var/lib/vmware/configstore/backup/current-store-1")).exists():
            self._configstore = parse_config_store(path)

    def check_compatible(self) -> None:
        # NOTE: Unable to use OS specific functions here, as this method can be called in ESXiPlugin.create
        if not self._configstore:
            raise UnsupportedPluginError("ESXi configstore not found on target")

    @internal
    def get(self, key: str, default: Any = None) -> dict[str, Any]:
        """Get configstore value for the specified key."""
        return self._configstore.get(key, default)


def parse_config_store(path: Path) -> dict[str, Any]:
    with SQLite3(path) as db:
        store = {}

        if table := db.table("Config"):
            for row in table.rows():
                component_name = row.Component
                config_group_name = row.ConfigGroup
                value_group_name = row.Name
                identifier_name = row.Identifier

                if component_name not in store:
                    store[component_name] = {}
                component = store[component_name]

                if config_group_name not in component:
                    component[config_group_name] = {}
                config_group = component[config_group_name]

                if value_group_name not in config_group:
                    config_group[value_group_name] = {}
                value_group = config_group[value_group_name]

                if identifier_name not in value_group:
                    value_group[identifier_name] = {}
                identifier = value_group[identifier_name]

                identifier["modified_time"] = row.ModifiedTime
                identifier["creation_time"] = row.CreationTime
                identifier["version"] = row.Version
                identifier["success"] = row.Success
                identifier["auto_conf_value"] = jsonlib.loads(row.AutoConfValue) if row.AutoConfValue else None
                identifier["user_value"] = jsonlib.loads(row.UserValue) if row.UserValue else None
                identifier["vital_value"] = jsonlib.loads(row.VitalValue) if row.VitalValue else None
                identifier["cached_value"] = jsonlib.loads(row.CachedValue) if row.CachedValue else None
                identifier["desired_value"] = jsonlib.loads(row.DesiredValue) if row.DesiredValue else None
                identifier["revision"] = row.Revision

        return store



#!/usr/bin/env python
from __future__ import annotations

import argparse
import logging
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dissect.cstruct.utils import hexdump

from dissect.target.containers.hdd import HddContainer
from dissect.target.exceptions import TargetError
from dissect.target.helpers.logging import get_logger
from dissect.target.plugin import arg
from dissect.target.tools.shell import python_shell
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    open_target,
    process_generic_arguments,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from dissect.target.container import Container
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import DiskCollection, FilesystemCollection, Target, VolumeCollection
    from dissect.target.volume import Volume

log = get_logger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


def indent(value: str, width: int = 1) -> None:
    """Indent and print value"""
    print(textwrap.indent(str(value), " " * (width * 4)))


def scope(*args, **kwargs) -> Callable[..., Any]:
    def decorator(obj: Callable[..., Any]) -> Callable[..., Any]:
        obj.__scope__ = args
        return obj

    return decorator


def _apply_scope(t: Target, scope: str) -> None:
    if scope == "disk":
        log.info("Only loading disks")
        return

    if scope == "volume":
        log.info("Only loading disks and volumes")
        t.disks.apply()
        t.volumes.apply(filesystems=False)
        return

    if scope == "filesystem":
        log.info("Only loading disks, volumes and filesystems")
        t.disks.apply()
        t.volumes.apply()
        t.filesystems.apply()
        return

    raise ValueError(f"Unknown scope: {scope}")


def _get_scope(t: Target, scope: str) -> DiskCollection | VolumeCollection | FilesystemCollection:
    if scope == "disk":
        return t.disks
    if scope == "volume":
        return t.volumes
    if scope == "filesystem":
        return t.filesystems

    raise ValueError(f"Unknown scope: {scope}")


def _get_obj(t: Target, args: argparse.Namespace) -> Iterator[Container | Volume | Filesystem]:
    entries = _get_scope(t, args.scope)

    if args.index is None:
        yield from entries
    elif args.index >= len(entries):
        raise IndexError(f"{args.scope.capitalize()} index {args.index} out of range (max {len(entries) - 1})")
    else:
        yield entries[args.index]


@scope("disk", "volume")
@arg("-s", "--skip", type=int, default=0, help="skip offset bytes from the beginning")
@arg("-n", "--length", type=int, default=512, help="number of bytes to read")
def cmd_hexdump(t: Target, args: argparse.Namespace) -> None:
    """hexdump some bytes"""
    for obj in _get_obj(t, args):
        print(f"# {obj}")

        obj.seek(args.skip)
        hexdump(obj.read(args.length), offset=args.skip)

        print()


def _info_hdd(obj: HddContainer) -> None:
    print(f"path={obj.hdd.path}")
    print()

    print("# Storage data")
    for storage in obj.hdd.descriptor.storage_data.storages:
        print(f"start={storage.start} end={storage.end}")
        for image in storage.images:
            indent(f"guid={image.guid}")
            indent(f"file={image.file}")
            indent(f"type={image.type}")
            print()

    print("# Snapshots")
    print(f"top_guid={obj.hdd.descriptor.snapshots.top_guid}")
    for shot in obj.hdd.descriptor.snapshots.shots:
        indent(f"guid={shot.guid}")
        indent(f"parent={shot.parent}")
        print()

    print("# Stream info")
    for storage, hds in obj.stream.streams:
        print(f"start={storage.start} end={storage.end}")
        i = 1
        while hds.parent is not None:
            indent(hds.fh, i)
            i += 1
            hds = hds.parent


@scope("disk", "volume")
def cmd_info(t: Target, args: argparse.Namespace) -> None:
    """print information"""
    for obj in _get_obj(t, args):
        print(f"# {obj}")

        if isinstance(obj, HddContainer):
            _info_hdd(obj)


@scope("disk", "volume", "filesystem")
def cmd_list(t: Target, args: argparse.Namespace) -> None:
    """list available items"""
    entries = _get_scope(t, args.scope)

    print(f"{args.scope.capitalize()}s:")
    for i, entry in enumerate(entries):
        print(f"[{i}]: {entry}")


@scope("disk", "volume", "filesystem")
def cmd_shell(t: Target, args: argparse.Namespace) -> None:
    """open (I)Python shell"""
    try:
        python_shell([t])
    except TargetError as e:
        log.error("Error opening shell: %s", e)  # noqa: TRY400
        log.debug("", exc_info=e)


def _add_cmds(parser: argparse.ArgumentParser, scope: str) -> None:
    sub = parser.add_subparsers(dest="cmd", required=False, help="available commands")

    for name, obj in globals().items():
        if not name.startswith("cmd_") or not callable(obj) or scope not in getattr(obj, "__scope__", ()):
            continue

        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument(
            "-i", "--index", type=int, default=None, help="index of the item to inspect", metavar="INDEX"
        )
        for args, kwargs in getattr(obj, "__args__", []):
            cmd_parser.add_argument(*args, **kwargs)
        cmd_parser.set_defaults(handler=obj)

        sub.add_parser(name[4:], help=obj.__doc__, parents=[cmd_parser])


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    scope = parser.add_subparsers(dest="scope", required=True, help="scope to inspect")

    base = argparse.ArgumentParser(add_help=False)
    base.add_argument("target", type=Path, help="Target to load", metavar="TARGET")
    configure_generic_arguments(base)

    parser_disk = scope.add_parser("disk", help="inspect disks", parents=[base])
    _add_cmds(parser_disk, "disk")

    parser_volume = scope.add_parser("volume", help="inspect volumes", parents=[base])
    _add_cmds(parser_volume, "volume")

    parser_filesystem = scope.add_parser("filesystem", help="inspect filesystems", parents=[base])
    _add_cmds(parser_filesystem, "filesystem")

    args, _ = parser.parse_known_args()
    process_generic_arguments(parser, args)

    try:
        t = open_target(args, apply=False)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    _apply_scope(t, args.scope)

    if args.index is not None and args.index >= len(_get_scope(t, args.scope)):
        print(f"{args.scope.capitalize()} index {args.index} out of range")
        print()
        cmd_list(t, args)
        return 1

    if args.cmd is None:
        cmd_list(t, args)
    else:
        args.handler(t, args)

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.editor.windowsnotepad import (
    WindowsNotepadPlugin,
    WindowsNotepadTab,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    import pytest

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target

text1 = "This is an unsaved tab, UTF-8 encoded with Windows (CRLF). It's only 88 characters long."
text2 = "Dissect は、インシデント対応のための優れたフレームワークです。 The Notepad window shows UTF-8 as the encoding. This text has 113 characters."  # noqa: E501
text3 = "This is a very short text."
text4 = "This is another short test. And we should be able to parse this."
text5 = "This is a test and the text is longer than 256 bytes. "
text6 = "This is a test and the text is longer than 65536 bytes. "
text7 = "This a text, which is nothing special. But I am going to modify it a bit. For example, I have removed quote some stuff. Adding a word in the beginning now... At this point, I've edited it quite a lot."  # noqa: E501
text8 = "Closing application now. It's saved but now I'm adding unsaved changes and closing the application again. Dit a few deletions!"  # noqa: E501
loremipsum = "Lorem ipsum dolor sit amet. Eum error blanditiis eum pariatur delectus ut consequuntur officiis a excepturi dignissimos et doloribus quia 33 perspiciatis soluta nam perspiciatis dolor. Ut repudiandae quidem cum sint modi qui sint consequatur. Aut autem quidem eum enim consequatur qui voluptate consequatur non similique voluptate. A vitae modi vel sint provident ut galisum tenetur sit voluptatem amet. Est impedit perspiciatis est repudiandae voluptates ut fugit alias! Eum magni esse aut velit illum qui excepturi aperiam. Ex dolores asperiores ut debitis omnis qui consequuntur dolore. Est voluptatem mollitia et quibusdam unde ea accusamus fuga. Cum quis galisum et impedit sunt qui aliquam perspiciatis sed modi quidem qui nisi molestias. Aut temporibus architecto ut neque voluptatem et consequatur deleniti sed accusantium quibusdam et omnis dignissimos ad rerum ipsam et rerum quia. Ut nihil repellat et eaque molestias quo iusto ipsum At optio sint eos quidem earum?\r\rEx deleniti unde eum tenetur rerum ea dolore numquam? Eos aperiam officiis et neque explicabo et enim atque ut eaque omnis non illum eveniet est molestias itaque et ratione voluptatem. Ea deserunt nemo et quos tempora et nostrum aperiam sit necessitatibus illo sit culpa placeat. Vel tempore quibusdam ut velit voluptate aut odio facere non voluptas earum est odio galisum et voluptas harum. Et blanditiis sapiente et nostrum laborum aut voluptatem explicabo a quasi assumenda. Est voluptatem quia eum minima galisum quo totam excepturi aut facilis enim vel voluptate repudiandae sit distinctio laboriosam. Quo possimus molestiae et molestiae accusantium est voluptas omnis sed obcaecati natus. Non vitae asperiores qui nostrum enim id saepe fugiat et incidunt quasi.\r\rEos ipsa facilis aut excepturi voluptatem a omnis magni vel magni iste. Sed ipsum consequatur qui reprehenderit deleniti et soluta molestiae. Ut vero assumenda id dolor ipsum in deleniti voluptatem aut quis quisquam sed repudiandae temporibus ab quia inventore. Sed velit fugit vel facere cumque et delectus ullam sed eaque impedit. Est veritatis dignissimos aut doloribus dolorem vel pariatur repellendus sit nesciunt similique eum architecto quia. Ea expedita veritatis eum dolorem molestiae ut enim fugit aut beatae quibusdam. Aut voluptas natus in quidem deleniti aut animi iure est incidunt tenetur qui culpa maiores! Et nostrum quaerat qui consequatur consequatur aut aliquam atque aut praesentium rerum et consequuntur exercitationem. Non accusantium ipsa vel consectetur vitae ut magnam autem et natus rerum ut consectetur inventore est doloremque temporibus 33 dolores doloribus! Aut perferendis optio et nostrum repellendus et fugit itaque ut nisi neque sed sint quaerat. Aut placeat architecto et eius sapiente eum molestiae quam. Quo mollitia sapiente non Quis neque non tempora laudantium. Quo distinctio quos et molestias natus sit veritatis consequuntur aut repellendus neque a porro galisum cum numquam nesciunt et animi earum? Aut dolorum dolore non assumenda omnis et molestiae amet id sint vero est eligendi harum sit temporibus magnam aut ipsam quos.\r\r"  # noqa: E501


def test_windows_tab_parsing() -> None:
    # Standalone parsing of tab files, so not using the plugin
    tab_file = Path(absolute_path("_data/plugins/apps/texteditor/windowsnotepad/unsaved-with-deletions.bin"))
    content = WindowsNotepadTab(tab_file)
    assert content.content == "Not saved aasdflasd"
    assert repr(content) == "<WindowsNotepadTab saved=False content_size=19 has_deleted_content=True>"


def test_windows_tab_plugin_deleted_contents(
    target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path, target_win_users: Target
) -> None:
    file_text_map = {
        "unsaved-with-deletions.bin": ("Not saved aasdflasd", "snUlltllafds tjkf"),
        "lots-of-deletions.bin": (
            "This a text, which is nothing special. But I am going to modify it a bit. "
            "For example, I have removed quote some stuff. "
            "Adding a word in the beginning now... "
            "At this point, I've edited it quite a lot.",
            "b a ,elpmac ydaerlae already thi laiceps emos",
        ),
    }

    tabcache = absolute_path("_data/plugins/apps/texteditor/windowsnotepad/")

    user = target_win_users.user_details.find(username="John")
    tab_dir = user.home_path.joinpath(
        "AppData/Local/Packages/Microsoft.WindowsNotepad_8wekyb3d8bbwe/LocalState/TabState"
    )

    fs_win.map_dir("Users\\John", tmp_path)

    for file in file_text_map:
        tab_file = str(tab_dir.joinpath(file))[3:]
        fs_win.map_file(tab_file, tabcache.joinpath(file))

    target_win.add_plugin(WindowsNotepadPlugin)

    records = list(target_win.windowsnotepad.tabs())

    # Check the amount of files
    assert len(list(tab_dir.iterdir())) == len(file_text_map.keys())
    assert len(records) == len(file_text_map.keys())

    # The recovered content in the records should match the original data, as well as the length
    for rec in records:
        assert rec.editor == "windowsnotepad"
        assert rec.content == file_text_map[rec.path.name][0]
        assert rec.deleted_content == file_text_map[rec.path.name][1]
        assert rec.source is not None


def test_windows_tab_plugin_default(
    target_win: Target,
    fs_win: VirtualFilesystem,
    tmp_path: Path,
    target_win_users: Target,
    caplog: pytest.LogCaptureFixture,
) -> None:
    file_text_map = {
        "c515e86f-08b3-4d76-844a-cddfcd43fcbb.bin": (text1, None),
        "85167c9d-aac2-4469-ae44-db5dccf8f7f4.bin": (text2, None),
        "dae80df8-e1e5-4996-87fe-b453f63fcb19.bin": (text3, "THis is "),
        "3f915e17-cf6c-462b-9bd1-2f23314cb979.bin": (text4, None),
        "ba291ccd-f1c3-4ca8-949c-c01f6633789d.bin": ((text5 * 5), None),
        "e609218e-94f2-45fa-84e2-f29df2190b26.bin": ((text6 * 1260), None),
        "3d0cc86e-dfc9-4f16-b74a-918c2c24188c.bin": (loremipsum, None),
        "wrong-checksum.bin": (text4, None),  # only added to check for corrupt checksum, not validity
        "cfe38135-9dca-4480-944f-d5ea0e1e589f.bin": (
            (loremipsum * 37)[:-2],
            None,
        ),  # removed the two newlines in this file
        "saved.bin": ("Saved!", None),
        "unsaved.bin": ("Not saved at all", "snUllt"),
        "unsaved-with-deletions.bin": ("Not saved aasdflasd", "snUlltllafds tjkf"),
        "lots-of-deletions.bin": (text7, "b a ,elpmac ydaerlae already thi laiceps emos"),
        "appclosed_saved_and_deletions.bin": (text8, None),
        "appclosed_unsaved.bin": ("Closing application now", None),
        "new-format.bin": ("", None),
        "stored_unsaved_with_new_data.bin": ("Stored to disk but unsaved, but with extra data.", None),
    }

    tabcache = absolute_path("_data/plugins/apps/texteditor/windowsnotepad/")

    user = target_win_users.user_details.find(username="John")
    tab_dir = user.home_path.joinpath(
        "AppData/Local/Packages/Microsoft.WindowsNotepad_8wekyb3d8bbwe/LocalState/TabState"
    )

    fs_win.map_dir("Users\\John", tmp_path)

    for file in file_text_map:
        tab_file = str(tab_dir.joinpath(file))[3:]
        fs_win.map_file(tab_file, tabcache.joinpath(file))

    target_win.add_plugin(WindowsNotepadPlugin)

    records = list(target_win.windowsnotepad.tabs())

    # Check the amount of files
    assert len(list(tab_dir.iterdir())) == len(file_text_map.keys())
    assert len(records) == len(file_text_map.keys())

    for line in caplog.text.split("\n"):
        # One file should still return contents, but there should be an entry for in the logging for a CRC missmatch.
        assert (
            "CRC32 mismatch in single-block file: wrong-checksum.bin (expected=deadbeef, actual=a48d30a6)" in line
            or "CRC32 mismatch" not in line
        )

    # The recovered content in the records should match the original data, as well as the length
    for rec in records:
        assert rec.editor == "windowsnotepad"
        assert rec.content == file_text_map[rec.path.name][0]
        assert rec.deleted_content == file_text_map[rec.path.name][1]
        assert rec.source is not None


def test_windows_saved_tab_plugin_extra_fields(
    target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path, target_win_users: Target
) -> None:
    file_text_map = {
        "saved.bin": (
            "Saved!",
            "C:\\Users\\user\\Desktop\\Saved!.txt",
            datetime(2024, 3, 28, 13, 7, 55, 482183, tzinfo=timezone.utc),
            "ed9b760289e614c9dc8776e7280abe870be0a85019a32220b35acc54c0ecfbc1",
        ),
        "appclosed_saved_and_deletions.bin": (
            text8,
            "C:\\Users\\user\\Desktop\\Saved.txt",
            datetime(2024, 3, 28, 13, 16, 21, 158279, tzinfo=timezone.utc),
            "8d0533144aa42e2d81e7474332bdef6473e42b699041528d55a62e5391e914ce",
        ),
    }

    tabcache = absolute_path("_data/plugins/apps/texteditor/windowsnotepad/")

    user = target_win_users.user_details.find(username="John")
    tab_dir = user.home_path.joinpath(
        "AppData/Local/Packages/Microsoft.WindowsNotepad_8wekyb3d8bbwe/LocalState/TabState"
    )

    fs_win.map_dir("Users\\John", tmp_path)

    for file in file_text_map:
        tab_file = str(tab_dir.joinpath(file))[3:]
        fs_win.map_file(tab_file, tabcache.joinpath(file))

    target_win.add_plugin(WindowsNotepadPlugin)

    records = list(target_win.windowsnotepad.tabs())

    # Check the amount of files
    assert len(list(tab_dir.iterdir())) == len(file_text_map.keys())
    assert len(records) == len(file_text_map.keys())

    # The recovered content in the records should match the original data, as well as the length and all the
    # other saved metadata
    for rec in records:
        assert rec.editor == "windowsnotepad"
        assert len(rec.content) == len(file_text_map[rec.path.name][0])
        assert rec.content == file_text_map[rec.path.name][0]
        assert rec.saved_path == file_text_map[rec.path.name][1]
        assert rec.ts == file_text_map[rec.path.name][2]
        assert rec.digest.sha256 == file_text_map[rec.path.name][3]
        assert rec.source is not None



from __future__ import annotations

import socket
import ssl
from pathlib import Path
from struct import unpack
from typing import TYPE_CHECKING
from unittest.mock import call, patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.remote import RemoteLoader, RemoteStream, RemoteStreamConnection
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Callable


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``RemoteLoader`` when opening a ``Target``."""
    path = "remote://127.0.0.1:1337"
    with (
        patch.object(ssl, "SSLContext", autospec=True),
        patch("dissect.target.loaders.remote.RemoteLoader.map"),
        patch("dissect.target.target.Target.apply"),
    ):
        target = opener(path)
        assert isinstance(target._loader, RemoteLoader)
        assert target.path == Path("127.0.0.1:1337")


def test_loader() -> None:
    """Test that ``RemoteLoader`` is correctly selected."""
    with patch.object(ssl, "SSLContext", autospec=True):
        loader = loader_open("remote://127.0.0.1:1337")
        assert isinstance(loader, RemoteLoader)


def test_stream() -> None:
    """Test that we can create a ``RemoteStream`` with a ``RemoteStreamConnection``."""

    with (
        patch.object(ssl, "SSLContext", autospec=True) as mock_ssl_context,
        patch.object(socket, "socket", autospec=True) as mock_socket,
    ):
        rsc = RemoteStreamConnection("remote://127.0.0.1", 9001, options={"ca": "A", "key": "B", "crt": "C"})
        assert rsc.is_connected() is False
        rsc.connect()

        rsc.remote_disk_data = b"ABC"
        rsc.remote_disk_response = None

        def send(data: bytes) -> None:
            _, offset, read = unpack(">BQQ", data)
            rsc.remote_disk_response = rsc.remote_disk_data[offset : offset + read]
            return len(data)

        def receive(num: bytes) -> bytes:
            return rsc.remote_disk_response

        rsc._ssl_sock.send = send
        rsc._ssl_sock.recv = receive
        rs = RemoteStream(rsc, 0, 3)
        rs.align = 1
        rs.seek(1)
        data = rs.read(2)
        rs.close()

        mock_socket.assert_called_with(socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_STREAM)

        expected = [
            call(ssl.PROTOCOL_TLSv1_2),
            call().load_default_certs(),
            call().load_cert_chain(certfile="C", keyfile="B"),
            call().load_verify_locations("A"),
            call().wrap_socket(rsc._socket, server_hostname="remote://127.0.0.1"),
            call().wrap_socket().connect(("remote://127.0.0.1", 9001)),
        ]

        assert data == b"BC"
        assert mock_ssl_context.mock_calls == expected
        assert rs.tell() == 3
        assert rsc.is_connected() is True


def test_stream_embedded() -> None:
    with (
        patch.object(ssl, "SSLContext", autospec=False) as mock_ssl_context,
        patch.object(socket, "socket", autospec=True),
    ):
        RemoteStreamConnection.configure("K", "C")
        RemoteStreamConnection("remote://127.0.0.1", 9001)
        mock_ssl_context.assert_has_calls([call().load_cert_chain_str(certfile="C", keyfile="K")])
